#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from os import PathLike
from pathlib import Path
from typing import Callable, Dict, Final, Iterator, List, Optional, Tuple, Union, cast
from warnings import warn

import numpy as np
import pydicom
from numpy import ndarray
from pydicom import DataElement, FileDataset
from pydicom.encaps import encapsulate
from pydicom.pixels.processing import apply_rescale, apply_voi_lut
from pydicom.uid import UID, ExplicitVRLittleEndian, ImplicitVRLittleEndian, JPEG2000TransferSyntaxes
from registry import bind_relevant_kwargs

from .basic_offset_table import BasicOffsetTable
from .logging import logger
from .tags import Tag
from .types import Dicom, PhotometricInterpretation, iterate_shared_functional_groups
from .volume import KeepVolume, VolumeHandler


try:
    # NOTE: This module will be open-sourced in the future
    import pynvjpeg  # type: ignore
except Exception:
    pynvjpeg = None


# Taken from https://pydicom.github.io/pydicom/dev/old/image_data_handlers.html
TransferSyntaxUIDs: Final[Dict[str, str]] = {
    "1.2.840.10008.1.2.1": "Explicit VR Little Endian",
    "1.2.840.10008.1.2": "Implicit VR Little Endian",
    "1.2.840.10008.1.2.2": "Explicit VR Big Endian",
    "1.2.840.10008.1.2.1.99": "Deflated Explicit VR Little Endian",
    "1.2.840.10008.1.2.5": "RLE Lossless",
    "1.2.840.10008.1.2.4.50": "JPEG Baseline (Process 1)",
    "1.2.840.10008.1.2.4.51": "JPEG Extended (Process 2 and 4)",
    "1.2.840.10008.1.2.4.57": "JPEG Lossless (Process 14)",
    "1.2.840.10008.1.2.4.70": "JPEG Lossless (Process 14, SV1)",
    "1.2.840.10008.1.2.4.80": "JPEG LS Lossless",
    "1.2.840.10008.1.2.4.81": "JPEG LS Lossy",
    "1.2.840.10008.1.2.4.90": "JPEG2000 Lossless",
    "1.2.840.10008.1.2.4.91": "JPEG2000",
    "1.2.840.10008.1.2.4.92": "JPEG2000 Multi-component Lossless",
    "1.2.840.10008.1.2.4.93": "JPEG2000 Multi-component",
}


# Pillow is relatively slow so we want to make sure that other handlers are used instead
default_data_handlers: List[Callable] = pydicom.config.pixel_data_handlers  # type: ignore
data_handlers = [h for h in default_data_handlers if "pillow" not in h.__name__]
pydicom.config.pixel_data_handlers = data_handlers  # type: ignore
assert (a := len(data_handlers) + 1) == (b := len(default_data_handlers)), f"Unexpected data handlers ({a} != {b})"


class NoImageError(Exception):
    pass


def is_native_byteorder(arr: ndarray) -> bool:
    r"""Checks if a numpy array has native byte order (Endianness)"""
    array_order = arr.dtype.byteorder
    if array_order in ["=", "|"]:
        return True
    return sys.byteorder == "little" and array_order == "<" or array_order == ">"


def is_inverted(photo_interp: str) -> bool:
    """Checks if pixel value 0 corresponds to white. See DICOM specification for more details."""
    warn(
        "is_inverted is deprecated. Please use PhotometricInterpretation.is_inverted",
        DeprecationWarning,
    )
    return PhotometricInterpretation.from_str(photo_interp).is_inverted


def invert_color(img: ndarray) -> ndarray:
    """The maximum value will become the minimum and vice versa"""
    return np.max(img) - img


def has_dicm_prefix(filename: Union[str, PathLike]) -> bool:
    """DICOM files have a 128 byte preamble followed by bytes 'DICM'."""
    with open(filename, "rb") as f:
        f.seek(128)
        return f.read(4) == b"DICM"


def is_compressed(dcm: Dicom) -> bool:
    r"""Checks if a DICOM is using a compressed transfer syntax"""
    syntax = dcm.file_meta.TransferSyntaxUID
    return syntax.is_compressed


def convert_frame_voi_lut(dcm: Dicom) -> Dicom:
    r"""Copies frame VOILUT information into the top level of a Dicom object."""
    for ds in iterate_shared_functional_groups(dcm):
        if Tag.FrameVOILUTSequence in ds:
            # set VOILUTSequence
            value = ds[Tag.FrameVOILUTSequence].value
            elem = DataElement(Tag.VOILUTSequence, "SQ", value)
            dcm[Tag.VOILUTSequence] = elem

            # set window tags
            window_c = value[0].WindowCenter
            window_w = value[0].WindowWidth
            dcm[Tag.WindowCenter] = DataElement(Tag.WindowCenter, "DS", window_c)
            dcm[Tag.WindowWidth] = DataElement(Tag.WindowWidth, "DS", window_w)
    return dcm


def strict_dcm_to_pixels(dcm: Dicom, dims: Tuple[int, ...], voi_lut: bool = True, rescale: bool = True) -> ndarray:
    """
    Interpret pixel data according to the TransferSyntaxUID stored in the DICOM dataset object.

    Args:
        dcm:
            DICOM object with pixel data
        dims:
            Tuple containing expected image shape
        voi_lut:
            Whether to apply VOI LUT transformation
        rescale:
            Whether to apply rescaling given in metadata

    Returns:
        Numpy ndarray of pixel data
    """
    try:
        pixels = apply_voi_lut(dcm.pixel_array, dcm) if voi_lut else dcm.pixel_array
        pixels = apply_rescale(pixels, dcm) if rescale else pixels
    except Exception:
        pixels = dcm.pixel_array
    return pixels.reshape(dims)


def loose_dcm_to_pixels(dcm: Dicom, dims: Tuple[int, ...], *args, **kwargs) -> ndarray:
    """
    Try all supported TransferSyntaxUIDs until one succeeds.
    Some mammograms have a mismatch between the TransferSyntaxUID and how the pixel data is actually encoded.

    Args:
        dcm:
            DICOM object with pixel data
        dims:
            Tuple containing expected image shape

    Raises:
        AttributeError:
            If the DICOM has no pixel data
        ValueError:
            If the DICOM has data that cannot be interpreted according to any supported TransferSyntaxUID

    Returns:
        Numpy ndarray of pixel data
    """
    for transfer_syntax_uid in TransferSyntaxUIDs.keys():
        try:
            dcm.file_meta.TransferSyntaxUID = UID(transfer_syntax_uid)
            pixels = strict_dcm_to_pixels(dcm, dims, *args, **kwargs)
            logger.warning(
                f"Able to parse pixels according to '{dcm.file_meta.TransferSyntaxUID}' "
                f"({TransferSyntaxUIDs[dcm.file_meta.TransferSyntaxUID]})"
            )
            return pixels
        # If the DICOM has no pixel data we can stop and raise an error
        except AttributeError as e:
            if "no pixel data to decode" in str(e):
                raise
        # Otherwise don't do anything, just see if the next TransferSyntaxUID works.
        except Exception:
            pass
    raise ValueError("Unable to parse the pixel array after trying all possible TransferSyntaxUIDs.")


def dcm_to_pixels(dcm: Dicom, dims: Tuple[int, ...], strict_interp: bool, *args, **kwargs) -> ndarray:
    """
    Try to parse pixel data according to a conformant interpretation,
    and if that fails then try to parse according to an alternative method if strict_interp==False.

    Args:
        dcm:
            DICOM object with pixel data
        dims:
            Tuple containing expected image shape
        strict_interp:
            If true, don't make any assumptions for trying to work around parsing errors

    Returns:
        Numpy ndarray of pixel data
    """
    try:
        return strict_dcm_to_pixels(dcm, dims, *args, **kwargs)
    except Exception as e:
        msg = (
            f"TransferSyntaxUID (0002, 0010) '{dcm.file_meta.TransferSyntaxUID}' "
            f"({TransferSyntaxUIDs[dcm.file_meta.TransferSyntaxUID]}) "
            f"does not appear to be correct. pydicom raised this error: '{e}'"
        )
        if strict_interp:
            raise ValueError(msg)
        logger.warning(msg)
        return loose_dcm_to_pixels(dcm, dims, *args, **kwargs)


def read_dicom_image(
    dcm: Dicom,
    stop_before_pixels: bool = False,
    override_shape: Optional[Tuple[int, ...]] = None,
    strict_interp: bool = False,
    volume_handler: VolumeHandler = KeepVolume(),
    as_uint8: bool = False,
    use_nvjpeg: Optional[bool] = None,
    nvjpeg_batch_size: Optional[int] = None,
    voi_lut: bool = True,
    inversion: bool = True,
    rescale: bool = True,
    convert_voi_lut: bool = True,
) -> ndarray:
    r"""
    Reads image data from an open DICOM file into a numpy array.

    Args:
        dcm:
            DICOM object to load images from
        stop_before_pixels:
            If ``True``, return randomly generated data
        shape:
            Manual shape override when ``stop_before_pixels`` is true. Should not include a channel dimension
        strict_interp:
            If ``True``, don't make any assumptions for trying to work around parsing errors
        volume_handler:
            Handler for processing 3D volumes
        as_uint8:
            If ``True``, convert non-uint8 outputs to uint8 using min/max normalization
        use_nvjpeg:
            If ``True``, decompress JPEG2000 images via GPU
        nvjpeg_batch_size:
            Batch size for GPU JPEG2000 decompression
        voi_lut:
            Whether to apply VOI LUT transformation
        inversion:
            Whether to apply correction for pixel value inversion
        rescale:
            Whether to apply rescaling given in metadata
        convert_voi_lut:
            Whether to convert frame VOI LUT to top level VOI LUT

    Shape:
        - Output: :math:`(C, H, W)` or :math:`(C, D, H, W)`
    """
    # some dicoms dont have any image data - raise NoImageError
    for necessary_field in ["Rows", "PhotometricInterpretation"]:
        if override_shape is None and not hasattr(dcm, necessary_field):
            raise NoImageError()
    pm = PhotometricInterpretation.from_dicom(dcm)

    C = pm.num_channels
    if override_shape is None:
        # If NumberOfFrames is 1 or not defined, we treat the DICOM image as a single channel 2D image (i.e. 1xHxW).
        # If NumberOfFrames is greater than 1, we treat the DICOM image as a single channel 3D image (i.e. 1xDxHxW).
        D, H, W = [int(v) for v in [dcm.get("NumberOfFrames", 1), dcm.Rows, dcm.Columns]]
        dims = (C, D, H, W) if D > 1 else (C, H, W)
    else:
        dims = (C,) + override_shape

    assert dims[0] in (1, 3), "channel dim == 1 or 3"
    assert 3 <= len(dims) <= 4, str(dims)

    # return random pixel data in correct shape when stop_before_pixels=True
    if stop_before_pixels or override_shape:
        return np.random.randint(0, 2**10, dims)

    # validation of compressed data
    if is_compressed(dcm) and (num_frames := dcm.get("NumberOfFrames", 1)) > 1:
        if BasicOffsetTable.is_present(dcm):
            bot = BasicOffsetTable.from_stream(dcm)
            if not (bot.is_valid and bot.num_frames == num_frames):
                if strict_interp:
                    raise ValueError(
                        "DICOM does not appear to have a valid basic offset table:\n"
                        f"NumberOfFrames: {dcm.get('NumberOfFrames', 1)}\n"
                        f"Offsets: {list(bot)}"
                    )
                else:
                    # replace the invalid BOT with an empty one
                    dcm.PixelData = BasicOffsetTable.remove_from(dcm)
                    bot = BasicOffsetTable.default()
                    dcm.PixelData = bot.prepend_to(dcm.PixelData)
            assert bot.is_valid

    # apply volume handling for 3D data
    if len(dims) == 4:
        if not isinstance(volume_handler, VolumeHandler):
            raise TypeError(f"volume_handler must be a VolumeHandler, not {type(volume_handler)}")
        volume_handler = cast(
            VolumeHandler,
            bind_relevant_kwargs(
                cast(Callable, volume_handler),
                use_nvjpeg=use_nvjpeg,
                batch_size=nvjpeg_batch_size,
            ),
        )
        dcm = volume_handler(dcm)
        D: int = int(dcm.get("NumberOfFrames", 1))
        dims = (C, D, *dims[-2:]) if D > 1 else (C, *dims[-2:])

    # Decompress with GPU if requested. ReduceVolume will decompress within the handler.
    if use_nvjpeg is None or use_nvjpeg:
        dcm = decompress(dcm, use_nvjpeg=use_nvjpeg, batch_size=nvjpeg_batch_size)

    # Convert frame VOI LUT to top level VOI LUT if requested
    if voi_lut and convert_voi_lut and not hasattr(dcm, "WindowCenter") or not hasattr(dcm, "WindowWidth"):
        dcm = convert_frame_voi_lut(dcm)

    # DICOM is channels last, so permute dims
    channels_last_dims = *dims[1:], dims[0]
    pixels = dcm_to_pixels(dcm, channels_last_dims, strict_interp, voi_lut=voi_lut, rescale=rescale)
    pixels = np.moveaxis(pixels, -1, 0)
    assert tuple(pixels.shape) == dims

    # some dicoms have different endianness - convert to native byte order
    if not is_native_byteorder(pixels):
        pixels = pixels.byteswap()
        pixels = pixels.view(pixels.dtype.newbyteorder("="))
    assert is_native_byteorder(pixels)

    # numpy byte order needs to explicitly be native "=" for torch conversion
    if pixels.dtype.byteorder != "=":
        pixels = pixels.view(pixels.dtype.newbyteorder("="))

    # in some dicoms, pixel value of 0 indicates white
    if pm.is_inverted and inversion:
        pixels = invert_color(pixels)

    # convert to uint8 if requested
    if as_uint8 and pixels.dtype != np.uint8:
        max, min = pixels.max(), pixels.min()
        delta = (max - min).clip(min=1)
        pixels = (pixels - min) / delta
        pixels = (pixels * 255).astype(np.uint8)

    return pixels


def path_to_dicom_path_list(path: Path) -> List[Path]:
    if path.is_dir():
        return [f for f in path.iterdir() if has_dicm_prefix(f)]
    if path.is_file() and has_dicm_prefix(path):
        return [path]
    else:
        raise FileNotFoundError(path)


def path_to_dicoms(path: Path) -> Iterator[Dicom]:
    for source in path_to_dicom_path_list(path):
        try:
            yield pydicom.dcmread(source)
        except Exception as e:
            logger.info(e)


def set_pixels(dcm: FileDataset, arr: np.ndarray, syntax: UID) -> FileDataset:
    r"""Sets the pixels of a DICOM object from a numpy array, accounting for TransferSyntaxUID."""
    if syntax.is_compressed:
        # NOTE: dcm.file_meta.TransferSyntaxUID must be an uncompressed TSUID when calling dcm.compress().
        # swap in a temporary uncompressed TSUID to address this
        dcm.file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
        # encapsulate frames
        if int(dcm.NumberOfFrames) > 1:
            new_data = encapsulate([a.tobytes() for a in arr], has_bot=False)
        else:
            new_data = encapsulate([arr.tobytes()])
        # update dcm with compressed attributes
        dcm.PixelData = new_data
        dcm.compress(syntax)
        dcm.file_meta.TransferSyntaxUID = syntax
    else:
        dcm.PixelData = arr.tobytes()
    dcm.file_meta.TransferSyntaxUID = syntax
    return dcm


def image_is_uint16(dcm: Dicom) -> bool:
    return int(dcm.get("PixelRepresentation", 0)) == 0


def decompress(
    dcm: Dicom,
    strict: bool = False,
    use_nvjpeg: Optional[bool] = None,
    batch_size: Optional[int] = None,
    verbose: bool = False,
) -> Dicom:
    r"""Decompress pixel data of a DICOM object

    Args:
        dcm: DICOM object to decompress
        strict: Raise an error if a decompressed DICOM is given as input
        use_nvjpeg: Use GPU accelerated decompression. Set to ``None`` to use if available
        batch_size: GPU decompression batch size. Set to ``None`` to defer to environment variable
        ``NVJPEG2K_BATCH_SIZE``

    Returns:
        DICOM object with decompressed pixel data and ExplicitVRLittleEndian transfer syntax.
    """
    tsuid = dcm.file_meta.TransferSyntaxUID
    if not tsuid.is_compressed:
        if strict:
            raise RuntimeError(f"TransferSyntaxUID {tsuid} is already decompressed")
        else:
            return dcm

    use_nvjpeg = use_nvjpeg if use_nvjpeg is not None else nvjpeg2k_is_available()

    # The GPU NVJPEG decoder doesn't seem to work with int16 pixel data
    # But this hasn't been thoroughly investigated
    use_nvjpeg = use_nvjpeg and image_is_uint16(dcm) and is_jpeg2k(dcm)

    if use_nvjpeg:
        batch_size = batch_size or int(os.environ.get("NVJPEG2K_BATCH_SIZE", 1))
        try:
            pixels = nvjpeg_decompress(dcm, batch_size, verbose)
        except Exception as e:
            # fall back to CPU
            logger.warning(f"Failed to decompress with nvjpeg: {e}")
            pixels = dcm.pixel_array
    else:
        pixels = dcm.pixel_array
    assert isinstance(dcm, FileDataset)
    dcm = set_pixels(dcm, pixels, ExplicitVRLittleEndian)
    return dcm


def nvjpeg2k_is_available() -> bool:  # pragma: no cover
    return pynvjpeg is not None


# TODO: support num_frames % batch_size != 0 in C++ extension
def _nvjpeg_get_batch_size(batch_size: int, num_frames: int) -> int:  # pragma: no cover
    while num_frames % batch_size != 0 and batch_size > 1:
        batch_size = batch_size - 1
    assert batch_size >= 1
    assert num_frames % batch_size == 0
    return batch_size


def is_jpeg2k(dcm: Dicom) -> bool:
    tsuid = dcm.file_meta.TransferSyntaxUID
    return tsuid in JPEG2000TransferSyntaxes


def nvjpeg_decompress(
    dcm: Dicom,
    batch_size: int = 4,
    verbose: bool = False,
) -> np.ndarray:  # pragma: no cover
    r"""Decompress pixel data of a DICOM object with NVJPEG2000

    Args:
        dcm: DICOM object to decompress
        batch_size: GPU decompression batch size. Set to ``None`` to defer to environment variable
        ``NVJPEG2K_BATCH_SIZE``.

    Returns:
        Decompressed pixel array
    """
    if not nvjpeg2k_is_available():
        raise ImportError('pynvjpeg is not available. Install with: pip install -e ".[j2k]"')
    elif not is_jpeg2k(dcm):
        raise ValueError(f"TransferSyntaxUID {dcm.file_meta.TransferSyntaxUID} is not supported for decompression")
    assert pynvjpeg is not None

    num_frames = dcm.get("NumberOfFrames", 1)

    # In case number of frames is present in "dcm" but stored as "None"
    dcm.NumberOfFrames = int(1 if num_frames is None else num_frames)
    rows = dcm.Rows
    cols = dcm.Columns

    # Build and validate a list of frames. Reject invalid JPEG2000 or signed JPEG2000
    # TODO: See if this can be done using only views into the original buffer. We are currently
    #   calling copy operations on the encoded frames.
    encoded_frames: List[bytes] = []
    for i, encoded_frame in enumerate(VolumeHandler.iterate_frames(dcm)):
        if not pynvjpeg.is_valid_jpeg2k(encoded_frame, len(encoded_frame)):
            raise ValueError(f"Frame {i} is not a valid JPEG2000 image")
        image_info = pynvjpeg.get_image_info_jpeg2k(encoded_frame, len(encoded_frame))
        for c in image_info.get("component_info", []):
            if c["sign"] != 0:
                raise ValueError(f"Signed images (frame {i}) are not supported")
        encoded_frames.append(encoded_frame)

    batch_size = _nvjpeg_get_batch_size(batch_size, num_frames)
    assert pynvjpeg is not None  # To fix a type error on the next line even though we already checked for this

    decoded_frames = pynvjpeg.decode_framelist_jpeg2k(
        b"".join(encoded_frames), [len(f) for f in encoded_frames], rows, cols, batch_size
    )

    # Squeeze extra dims for 2D inputs
    if num_frames == 1:
        H, W = decoded_frames.shape[-2:]
        decoded_frames = decoded_frames.reshape(H, W)

    return decoded_frames



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from pathlib import Path

from ..merge_series import merge_series


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", help="path to a folder with DICOM files", type=Path)
    return parser


def main(args: argparse.Namespace) -> None:
    merge_series(Path(args.path))


def entrypoint() -> None:
    main(get_parser().parse_args())


if __name__ == "__main__":
    entrypoint()



import hashlib
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Final, List, Optional

from pydicom import Dataset
from pydicom.dataset import PrivateBlock

from .tags import Tag


@dataclass
class MedCogElement:
    value: Any
    VR: str


@dataclass
class MedCogElementCreator:
    name: str
    create_element: Callable[[Dataset], MedCogElement]


class VR(Enum):
    other_byte_string = "OB"
    long_text = "LT"


def get_study_date(ds: Dataset) -> Optional[datetime]:
    if (study_date := ds.get(Tag.StudyDate, None)) and study_date.value:
        return datetime.strptime(study_date.value, DICOM_DATE_FORMAT)


def get_year(ds: Dataset) -> MedCogElement:
    # Elements of dates that are not permitted for disclosure include the day, month, and any other information that
    # is more specific than the year of an event. For instance, the date “January 1, 2009” could not be reported at
    # this level of detail. However, it could be reported in a de-identified data set as “2009”.
    # https://www.hhs.gov/hipaa/for-professionals/privacy/special-topics/de-identification/index.html
    study_date = get_study_date(ds)
    year_str = str(study_date.year) if study_date else "????"
    return MedCogElement(year_str, VR.long_text.value)


def hash_pixel_data(ds: Dataset) -> MedCogElement:
    if pixel_data := ds.get(Tag.PixelData, None):
        value = hash_any(pixel_data)
    else:
        value = b""
    return MedCogElement(value, VR.other_byte_string.value)


MEDCOG_ELEMENT_CREATORS: Final[List[MedCogElementCreator]] = [
    MedCogElementCreator("PixelDataHash", hash_pixel_data),
    MedCogElementCreator("StudyYear", get_year),
]

DICOM_DATE_FORMAT: Final[str] = r"%Y%m%d"
MEDCOG_ADDR: Final[int] = int("0x" + "".join(f"{ord(v):2x}" for v in "MC"), 16)
MEDCOG_NAME: Final[str] = "MedCognetics"
PRIVATE_ELEMENTS_DESCRIPTION: Final[str] = ",".join(f.name for f in MEDCOG_ELEMENT_CREATORS)
assert len(PRIVATE_ELEMENTS_DESCRIPTION) <= 10240, f"Description exceeds maximum length for VR '{VR.long_text.value}'"


def hash_bytes(x: bytes) -> bytes:
    sha256 = hashlib.sha256()
    sha256.update(x)
    return sha256.digest()


def hash_any(value: Any, encoding: str = "utf-8") -> bytes:
    # NOTE: Hashing patient names and other identifying fields may be unsafe
    # due to bute-force or rainbow table attacks.
    value = str(value).encode(encoding)
    return hash_bytes(value)


def get_medcog_elements(ds: Dataset) -> List[MedCogElement]:
    return [field.create_element(ds) for field in MEDCOG_ELEMENT_CREATORS]


def get_medcog_block(ds: Dataset, create: bool = False) -> PrivateBlock:
    return ds.private_block(MEDCOG_ADDR, MEDCOG_NAME, create=create)


def store_medcog_elements(ds: Dataset, field_values: List[MedCogElement]) -> None:
    block = get_medcog_block(ds, create=True)
    block.add_new(0, VR.long_text.value, PRIVATE_ELEMENTS_DESCRIPTION)
    for i, field in enumerate(field_values):
        block.add_new(i + 1, field.VR, field.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import runpy
import shutil
import sys
from pathlib import Path

import pytest


@pytest.fixture
def dicom_file():
    pydicom = pytest.importorskip("pydicom")
    return pydicom.data.get_testdata_file("CT_small.dcm")


def test_validate(dicom_file, capsys, tmp_path):
    dest = Path(tmp_path, "file.dcm")
    shutil.copy(dicom_file, dest)
    sys.argv = [
        sys.argv[0],
        str(dest),
        "--no-color",
    ]
    with pytest.raises(SystemExit):
        runpy.run_module("dicom_utils.cli.validate", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()

    lines = [re.sub("[^A-Za-z0-9\t]+", "", s).split("\t") for s in captured.out.split("\n")]

    assert lines[0] == ["Tag", "Priority", "State", "Message"]
    assert len(lines) == 34
    # TODO consider validating the full output


@pytest.mark.parametrize("no_color", [True, False])
@pytest.mark.parametrize("failing_only", [True, False])
def test_validate_args(dicom_file, capsys, tmp_path, no_color, failing_only):
    dest = Path(tmp_path, "file.dcm")
    shutil.copy(dicom_file, dest)
    sys.argv = [
        sys.argv[0],
        str(dest),
    ]
    if no_color:
        sys.argv.append("--no-color")
    if failing_only:
        sys.argv.append("--failing-only")

    with pytest.raises(SystemExit):
        runpy.run_module("dicom_utils.cli.validate", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert captured.out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import shutil
from dataclasses import fields, replace
from io import IOBase
from os import PathLike
from pathlib import Path
from typing import Any, List, Optional, cast

import pydicom
import pytest
from pydicom import DataElement, Sequence
from pydicom.dataset import Dataset
from pydicom.multival import MultiValue
from pydicom.uid import AllTransferSyntaxes, SecondaryCaptureImageStorage

from dicom_utils.container import FileRecord
from dicom_utils.container.record import (
    SOPUID,
    STANDARD_MAMMO_VIEWS,
    DicomFileRecord,
    DicomImageFileRecord,
    MammogramFileRecord,
    ModalityHelper,
    StandardizedFilename,
)
from dicom_utils.dicom_factory import DicomFactory
from dicom_utils.tags import Tag
from dicom_utils.types import Laterality, MammogramType, MammogramView, PixelSpacing, ViewPosition, get_value


class TestStandardizedFilename:
    @pytest.mark.parametrize(
        "inp, exp",
        [
            pytest.param(Path("foo_1.txt"), StandardizedFilename("foo_1.txt")),
            pytest.param(Path("bar_1234"), StandardizedFilename("bar_1234")),
            pytest.param(Path("foo.txt"), StandardizedFilename("foo_1.txt")),
        ],
    )
    def test_create(self, inp, exp):
        p = StandardizedFilename(inp)
        assert isinstance(p, StandardizedFilename)
        assert isinstance(p, Path)
        assert p == exp

    @pytest.mark.parametrize(
        "inp,exp",
        [
            pytest.param(StandardizedFilename("foo_1.txt"), "1"),
            pytest.param(StandardizedFilename("bar_2.txt"), "2"),
            pytest.param(StandardizedFilename("baz_3"), "3"),
            pytest.param(StandardizedFilename("bar_1234.txt"), "1234"),
            pytest.param(StandardizedFilename("bar_1234"), "1234"),
        ],
    )
    def test_file_id(self, inp, exp):
        assert inp.file_id == exp

    @pytest.mark.parametrize(
        "inp,id,exp",
        [
            pytest.param(StandardizedFilename("foo.txt"), "1", StandardizedFilename("foo_1.txt")),
            pytest.param(StandardizedFilename("foo_2.txt"), "1", StandardizedFilename("foo_1.txt")),
            pytest.param(StandardizedFilename("bar"), "1", StandardizedFilename("bar_1")),
            pytest.param(StandardizedFilename("bar_123.txt"), "234", StandardizedFilename("bar_234.txt")),
        ],
    )
    def test_with_file_id(self, inp, id, exp):
        assert inp.with_file_id(id) == exp

    @pytest.mark.parametrize(
        "inp,exp",
        [
            pytest.param(StandardizedFilename("ffdm_1.dcm"), "ffdm"),
            pytest.param(StandardizedFilename("ffdm_mag_spot_1.dcm"), "ffdm_mag_spot"),
            pytest.param(StandardizedFilename("ffdm_mag_spot_1"), "ffdm_mag_spot"),
        ],
    )
    def test_prefix(self, inp, exp):
        assert inp.prefix == exp

    @pytest.mark.parametrize(
        "inp,val,exp",
        [
            pytest.param(StandardizedFilename("file_1.dcm"), ["ffdm"], StandardizedFilename("ffdm_1.dcm")),
            pytest.param(StandardizedFilename("file_2.dcm"), ["ffdm", "spot"], StandardizedFilename("ffdm_spot_2.dcm")),
        ],
    )
    def test_with_prefix(self, inp, val, exp):
        assert inp.with_prefix(*val) == exp

    @pytest.mark.parametrize(
        "inp,mod,exp",
        [
            pytest.param(StandardizedFilename("ffdm_1.dcm"), "spot", StandardizedFilename("ffdm_spot_1.dcm")),
            pytest.param(StandardizedFilename("ffdm_mag_2.dcm"), "spot", StandardizedFilename("ffdm_mag_spot_2.dcm")),
        ],
    )
    def test_add_modifier(self, inp, mod, exp):
        assert inp.add_modifier(mod) == exp


class TestFileRecord:
    @pytest.fixture
    def record_factory(self, tmp_path):
        def func(filename: PathLike = Path("foo.txt")):
            filename = Path(tmp_path, filename)
            filename.parent.mkdir(exist_ok=True, parents=True)
            filename.touch()
            record = FileRecord(filename)
            return record

        return func

    def test_repr(self, record_factory):
        rec = record_factory("a.txt")
        s = repr(rec)
        assert isinstance(s, str)
        assert rec.__class__.__name__ in s

    def test_hash(self, record_factory):
        rec1 = record_factory("foo.txt")
        rec2 = record_factory("foo.txt")
        rec3 = record_factory("bar.txt")
        assert hash(rec1) == hash(rec2)
        assert hash(rec1) != hash(rec3)

    def test_eq(self, record_factory):
        rec1 = record_factory("foo.txt")
        rec2 = record_factory("foo.txt")
        rec3 = record_factory("bar.txt")
        assert rec1 == rec2
        assert rec1 != rec3

    def test_compare(self, record_factory):
        rec1 = record_factory("a.txt")
        rec2 = record_factory("b.txt")
        rec3 = record_factory("c.txt")
        assert sorted([rec3, rec2, rec1]) == [rec1, rec2, rec3]

    def test_has_uid(self, record_factory):
        rec = record_factory("a.txt")
        assert rec.has_uid

    def test_get_uid(self, record_factory):
        rec = record_factory("a.txt")
        assert rec.get_uid() == str(rec.path.stem)

    @pytest.mark.parametrize(
        "path,target,exp",
        [
            pytest.param("foo/bar.txt", "foo/", "bar.txt"),
            pytest.param("foo/bar/baz.txt", "foo/baz/bar.txt", "../../bar/baz.txt"),
            pytest.param("foo/bar/baz.txt", "foo/baz/", "../bar/baz.txt"),
        ],
    )
    def test_relative_to(self, tmp_path, path, target, exp, record_factory):
        target = Path(tmp_path, target)
        rec = record_factory(path)
        relative_rec = rec.relative_to(target)
        assert type(relative_rec) == type(rec)
        assert relative_rec.path == Path(exp)

    @pytest.mark.parametrize(
        "path,target,exp",
        [
            pytest.param("foo/bar.txt", "foo/baz.txt", True),
            pytest.param("foo/baz/bar.txt", "foo/baz.txt", False),
            pytest.param("foo/baz/bar.txt", "foo/bar/baz.txt", False),
            pytest.param("foo.txt", "foo.txt", True),
        ],
    )
    def shares_directory_with(self, path1, path2, exp, record_factory):
        rec1 = record_factory(path1)
        rec2 = record_factory(path2)
        assert rec1.shares_directory_with(rec2) == exp
        assert rec2.shares_directory_with(rec1) == exp

    def test_present_fields(self, record_factory):
        rec = record_factory()
        actual = dict(rec.present_fields())
        expected = {field.name: value for field in fields(rec) if (value := getattr(rec, field.name)) != field.default}
        assert actual == expected

    def test_file_size(self, record_factory):
        rec = record_factory()
        actual = rec.file_size
        expected = rec.path.stat().st_size
        assert actual == expected

    def test_is_compressed(self, record_factory):
        rec = record_factory()
        assert not rec.is_compressed

    @pytest.mark.parametrize(
        "path,file_id,exp",
        [
            pytest.param("foo/bar.txt", None, "bar_bar.txt"),
            pytest.param("foo/bar.txt", "2", "bar_2.txt"),
            pytest.param("foo/bar.txt", 1, "bar_1.txt"),
        ],
    )
    def test_standardized_filename(self, path, file_id, exp, record_factory):
        rec = record_factory(path)
        actual = rec.standardized_filename(file_id)
        assert isinstance(actual, StandardizedFilename)
        assert actual == Path(exp)

    def test_read(self, record_factory):
        rec = record_factory()
        stream = rec.__class__.read(rec.path)
        assert isinstance(stream, IOBase)

    @pytest.mark.parametrize(
        "path,symlink",
        [
            pytest.param("foo/bar/baz.txt", "foo/link/baz.txt"),
            pytest.param("1/2/baz.txt", "2/1/baz.txt"),
            pytest.param("1/2/baz.txt", "2/foo.txt"),
        ],
    )
    def test_to_symlink(self, tmp_path, path, symlink, record_factory):
        rec = record_factory(path)
        symlink = Path(tmp_path, symlink)
        symlink_rec = rec.to_symlink(symlink)
        assert symlink_rec.path.is_symlink()
        assert symlink_rec.path.resolve() == rec.path

    def test_to_dict(self, record_factory):
        rec = record_factory("a.txt")
        rec_dict = rec.to_dict()
        assert rec_dict["record_type"] == rec.__class__.__name__
        assert rec_dict["path"] == str(rec.path.absolute())
        assert rec_dict["resolved_path"] == str(rec.path.resolve().absolute())
        restored = FileRecord.from_dict(rec_dict)
        assert isinstance(restored, type(rec))
        assert isinstance(restored.path, Path)
        assert restored == rec


def make_view_modifier_code(meaning: str) -> Dataset:
    vc = Dataset()
    vc[Tag.CodeMeaning] = DataElement(Tag.CodeMeaning, "ST", meaning)
    return vc


class TestDicomFileRecord(TestFileRecord):
    @pytest.fixture
    def record_factory(self, tmp_path, dicom_file):
        def func(filename: PathLike = Path("foo.txt"), **kwargs):
            filename = Path(tmp_path, filename)
            filename.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(dicom_file, filename)
            record = DicomFileRecord.from_file(filename, **kwargs)
            return record

        return func

    def test_from_file_assigns_tags(self, record_factory):
        rec = record_factory()
        dcm = pydicom.dcmread(rec.path, stop_before_pixels=True)
        for field in fields(rec):
            tag = getattr(Tag, field.name, None)
            # shortcut for non-tag fields
            if tag is None:
                continue
            value = getattr(rec, field.name, None)
            expected = get_value(dcm, tag, None, try_file_meta=True)
            # shortcut for modality so subclassing tests that override modality
            # won't fail
            if tag == Tag.Modality:
                assert value is not None
            elif expected is not None:
                assert value == expected

    def test_from_file_only_reads_once(self, mocker, record_factory):
        spy = mocker.spy(pydicom, "dcmread")
        proto = record_factory()
        spy.reset_mock()
        proto.__class__.from_file(proto.path)
        spy.assert_called_once()

    def test_from_file_bad_suffix(self, tmp_path, mocker, record_factory):
        proto = record_factory()
        new_path = Path(f"{proto.path}.123")
        shutil.copy(proto.path, new_path)
        rec = proto.__class__.from_file(new_path)
        assert rec.__class__ == proto.__class__
        assert rec.path == new_path

    @pytest.mark.parametrize("file_exists", [True, False])
    def test_from_dicom(self, mocker, record_factory, file_exists):
        proto = record_factory()
        with pydicom.dcmread(proto.path) as dcm:
            if not file_exists:
                proto.path.unlink()
                assert not proto.path.is_file()
            result = proto.__class__.from_dicom(proto.path, dcm)
        assert result == proto

    @pytest.mark.parametrize(
        "sop,series,exp",
        [
            pytest.param("1.2.345", "2.3.456", True),
            pytest.param("1.2.345", None, True),
            pytest.param(None, "2.3.456", True),
            pytest.param(None, None, False),
        ],
    )
    def test_has_uid(self, sop, series, exp, record_factory):
        rec = record_factory()
        rec = rec.replace(
            SOPInstanceUID=sop,
            SeriesInstanceUID=series,
        )
        assert rec.has_uid == exp

    @pytest.mark.parametrize(
        "sop,series,prefer_sop,exp",
        [
            pytest.param("1.2.345", "2.3.456", False, "2.3.456"),
            pytest.param("1.2.345", "2.3.456", True, "1.2.345"),
            pytest.param("1.2.345", None, False, "1.2.345"),
            pytest.param(None, "2.3.456", True, "2.3.456"),
            pytest.param(None, None, None, None, marks=pytest.mark.xfail(raises=AttributeError)),
        ],
    )
    def test_get_uid(self, sop, series, prefer_sop, exp, record_factory):
        rec = record_factory()
        rec = rec.replace(
            SOPInstanceUID=sop,
            SeriesInstanceUID=series,
        )
        uid = rec.get_uid(prefer_sop)
        assert uid == exp

    @pytest.mark.parametrize(
        "path,modality,file_id,exp",
        [
            pytest.param("foo/bar.dcm", "CT", None, "ct_1-2-345.dcm"),
            pytest.param("foo/bar.dcm", "US", "2", "us_2.dcm"),
            pytest.param("foo/bar.dcm", "MG", 1, "mg_1.dcm"),
        ],
    )
    def test_standardized_filename(self, path, modality, file_id, exp, record_factory):
        rec = record_factory(path)
        sop = "1.2.345"
        rec = rec.replace(SOPInstanceUID=sop, Modality=modality)
        actual = rec.standardized_filename(file_id)
        assert isinstance(actual, StandardizedFilename)
        assert actual == Path(exp)

    def test_read(self, mocker, record_factory):
        spy = mocker.spy(pydicom, "dcmread")
        rec = record_factory()
        spy.reset_mock()
        stream = rec.__class__.read(rec.path)
        assert isinstance(stream, pydicom.FileDataset)
        spy.assert_called_once()
        assert spy.mock_calls[0].kwargs["stop_before_pixels"]

    @pytest.mark.parametrize(
        "study_date,exp",
        [
            pytest.param("20200101", 2020),
            pytest.param("00010101", 1),
            pytest.param("", None),
            pytest.param("fooo", None),
            pytest.param("10", None),
        ],
    )
    def test_year(self, study_date, exp, record_factory):
        rec = record_factory()
        rec = rec.replace(StudyDate=study_date)
        assert rec.year == exp

    def test_to_dict(self, record_factory):
        super().test_to_dict(record_factory)
        rec = record_factory("a.dcm")
        rec_dict = rec.to_dict()
        assert rec_dict["Modality"] == rec.Modality

    def test_hash(self, record_factory):
        rec1 = record_factory("a.dcm")
        rec2 = record_factory("b.dcm")
        rec3 = record_factory("a.dcm")
        rec1 = rec1.replace(SOPInstanceUID="123")
        rec2 = rec2.replace(SOPInstanceUID="123")
        assert hash(rec1) == hash(rec2)
        assert hash(rec1) != hash(rec3)

    @pytest.mark.parametrize("has_uid", [False, True])
    def test_eq(self, record_factory, has_uid):
        rec1 = record_factory("a.dcm")
        rec2 = record_factory("b.dcm")
        rec3 = record_factory("a.dcm")
        rec1 = rec1.replace(SOPInstanceUID="123" if has_uid else None)
        rec2 = rec2.replace(SOPInstanceUID="123" if has_uid else None)
        rec3 = rec3.replace(SOPInstanceUID="234" if has_uid else None)

        if has_uid:
            assert rec1 == rec2
            assert rec1 != rec3
        else:
            assert rec1 != rec2
            assert rec1 == rec3

    @pytest.mark.parametrize(
        "addr,name,ts,exp",
        [
            (None, None, None, None),
            ("foo", None, None, "foo"),
            (None, "foo", None, "foo"),
            (None, None, "foo", "foo"),
            ("foo", "bar", "baz", "foo"),
            (None, "bar", "baz", "baz"),
        ],
    )
    def test_site(self, addr, name, ts, exp, record_factory):
        rec = record_factory("foo.dcm")
        rec = rec.replace(
            InstitutionAddress=addr,
            InstitutionName=name,
            TreatmentSite=ts,
        )
        assert rec.site == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), False),
            (DicomFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            # Otherwise compare SOPInstanceUID
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                False,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                True,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                False,
            ),
        ],
    )
    def test_lt(self, rec1, rec2, exp):
        assert (rec1 < rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), True),
            (DicomFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            # Otherwise compare SOPInstanceUID
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                False,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                False,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                True,
            ),
        ],
    )
    def test_gt(self, rec1, rec2, exp):
        assert (rec1 > rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), False),
            (DicomFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            # Otherwise compare SOPInstanceUID
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                True,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                True,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                False,
            ),
        ],
    )
    def test_le(self, rec1, rec2, exp):
        assert (rec1 <= rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            (DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), True),
            (DicomFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            # Otherwise compare SOPInstanceUID
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                True,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                False,
            ),
            (
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                True,
            ),
        ],
    )
    def test_ge(self, rec1, rec2, exp):
        assert (rec1 >= rec2) == exp


class TestDicomImageFileRecord(TestDicomFileRecord):
    @pytest.fixture
    def record_factory(self, tmp_path, dicom_file):
        def func(
            filename: PathLike = Path("foo.txt"),
            Rows: Optional[int] = 128,
            Columns: Optional[int] = 128,
            NumberOfFrames: Optional[int] = None,
            Modality: Optional[str] = "CT",
            view_modifier_code: Optional[str] = None,
            PixelSpacing: Optional[Any] = MultiValue(str, [0.661468, 0.661468]),
            ImagerPixelSpacing: Optional[Any] = None,
            **kwargs,
        ):
            filename = Path(tmp_path, filename)
            filename.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(dicom_file, filename)
            record = DicomImageFileRecord.from_file(filename, **kwargs)

            if view_modifier_code is not None:
                vc = make_view_modifier_code(view_modifier_code)
                view_modifier_code_seq = Sequence([vc])
            else:
                view_modifier_code_seq = None

            record = record.replace(
                Rows=Rows,
                Columns=Columns,
                NumberOfFrames=NumberOfFrames,
                Modality=kwargs.get("modality", Modality),
                ViewModifierCodeSequence=view_modifier_code_seq,
                PixelSpacing=PixelSpacing,
                ImagerPixelSpacing=ImagerPixelSpacing,
            )
            return record

        return func

    @pytest.mark.parametrize(
        "rows,columns,nf,exp",
        [
            pytest.param(100, 100, 100, True),
            pytest.param(100, 100, None, True),
            pytest.param(None, 100, None, False),
            pytest.param(100, None, None, False),
            pytest.param(None, None, None, False),
        ],
    )
    def test_is_valid_image(self, rows, columns, nf, exp, record_factory):
        rec = record_factory(Rows=rows, Columns=columns, NumberOfFrames=nf)
        assert rec.is_valid_image == exp

    @pytest.mark.parametrize(
        "tsuid,exp",
        [pytest.param(tsuid, tsuid.is_compressed) for tsuid in AllTransferSyntaxes],
    )
    def test_is_compressed(self, tsuid, exp, record_factory):
        rec = record_factory()
        rec = rec.replace(TransferSyntaxUID=tsuid)
        assert rec.is_compressed == exp

    @pytest.mark.parametrize(
        "rows,columns,nf,exp",
        [
            pytest.param(100, 100, 100, True),
            pytest.param(100, 100, 1, False),
            pytest.param(100, 100, None, False),
            pytest.param(None, 100, 100, False),
            pytest.param(None, 100, None, False),
            pytest.param(100, None, None, False),
            pytest.param(None, None, None, False),
        ],
    )
    def test_is_volume(self, rows, columns, nf, exp, record_factory):
        rec = record_factory(Rows=rows, Columns=columns, NumberOfFrames=nf)
        assert rec.is_volume == exp

    @pytest.mark.parametrize(
        "code,exp",
        [
            pytest.param("spot compression", False),
            pytest.param("magnified", True),
            pytest.param("magnification", True),
        ],
    )
    def test_is_magnified(self, code, exp, record_factory):
        rec = record_factory(view_modifier_code=code)
        assert rec.is_magnified == exp

    @pytest.mark.parametrize(
        "spacing,imager_spacing,exp",
        [
            pytest.param("[0.01, 0.01]", None, PixelSpacing(0.01, 0.01)),
            pytest.param(None, "[0.01, 0.01]", PixelSpacing(0.01, 0.01)),
            pytest.param(MultiValue(str, [0.02, 0.02]), None, PixelSpacing(0.02, 0.02)),
            pytest.param(None, MultiValue(str, [0.02, 0.02]), PixelSpacing(0.02, 0.02)),
            pytest.param(None, None, None),
        ],
    )
    def test_pixel_spacing(self, spacing, imager_spacing, exp, record_factory):
        rec = record_factory(PixelSpacing=spacing, ImagerPixelSpacing=imager_spacing)
        assert rec.pixel_spacing == exp

    def test_to_dict(self, record_factory):
        super().test_to_dict(record_factory)
        rec = record_factory("a.txt", view_modifier_code="code")
        rec_dict = rec.to_dict()
        restored = FileRecord.from_dict(rec_dict)
        assert isinstance(restored, DicomImageFileRecord)
        assert list(restored.view_modifier_code_meanings) == ["code"]

    @pytest.mark.parametrize(
        "tag,val,exp",
        [
            pytest.param(Tag.StudyDescription, None, False),
            pytest.param(Tag.StudyDescription, "foo", False),
            pytest.param(Tag.StudyDescription, "XR Soft tissue specimen", True),
        ],
    )
    def test_is_specimen(self, tag, val, exp, record_factory):
        rec = record_factory()
        rec = rec.replace(**{tag.name: val})
        assert rec.is_specimen == exp


class TestMammogramFileRecord(TestDicomFileRecord):
    @pytest.fixture
    def record_factory(self, tmp_path, dicom_file):
        def func(
            filename: PathLike = Path("foo.dcm"),
            Rows: Optional[int] = 128,
            Columns: Optional[int] = 128,
            NumberOfFrames: Optional[int] = None,
            Modality: Optional[str] = "MG",
            view_modifier_code: Optional[str] = "medio-lateral oblique",
            PixelSpacing: Optional[Any] = MultiValue(str, [0.661468, 0.661468]),
            ImagerPixelSpacing: Optional[Any] = None,
            laterality: Optional[Laterality] = Laterality.LEFT,
            view_position: Optional[ViewPosition] = ViewPosition.MLO,
            mammogram_type: Optional[MammogramType] = MammogramType.FFDM,
            **kwargs,
        ):
            filename = Path(tmp_path, filename)
            filename.parent.mkdir(exist_ok=True, parents=True)
            with pydicom.dcmread(dicom_file) as dcm:
                dcm.Modality = "MG"
                dcm.save_as(filename)
            record = MammogramFileRecord.from_file(filename, **kwargs)

            if view_modifier_code is not None:
                vc = make_view_modifier_code(view_modifier_code)
                view_modifier_code_seq = Sequence([vc])
            else:
                view_modifier_code_seq = None

            if laterality is not None:
                record = record.replace(laterality=laterality)
            if view_position is not None:
                record = record.replace(view_position=view_position)
            if mammogram_type is not None:
                record = record.replace(mammogram_type=mammogram_type)

            record = record.replace(
                Rows=Rows,
                Columns=Columns,
                NumberOfFrames=NumberOfFrames,
                Modality=kwargs.get("modality", Modality),
                ViewModifierCodeSequence=view_modifier_code_seq,
                PixelSpacing=PixelSpacing,
                ImagerPixelSpacing=ImagerPixelSpacing,
            )
            return record

        return func

    @pytest.mark.parametrize(
        "dtype,attr",
        [
            pytest.param(Laterality, "laterality"),
            pytest.param(ViewPosition, "view_position"),
            pytest.param(MammogramType, "mammogram_type"),
        ],
    )
    def test_from_file_assigns_mammogram_attrs(self, mocker, dtype, attr, record_factory):
        m = mocker.patch.object(dtype, "from_dicom", spec_set=dtype)
        rec = record_factory(**{attr: None})
        m.assert_called_once()
        assert getattr(rec, attr) == m()

    @pytest.mark.parametrize(
        "paddle,code,view_pos,exp",
        [
            pytest.param("SPOT", None, None, True),
            pytest.param("SPOT COMPRESSION", None, None, True),
            pytest.param(None, None, None, False),
            pytest.param(None, "spot compression", None, True),
            pytest.param(None, None, "CCSpot", True),
            pytest.param("7.5CM SPTMAG", None, None, True),
        ],
    )
    def test_is_spot_compression(self, paddle, code, view_pos, exp, record_factory):
        record = record_factory()
        seq = Sequence([make_view_modifier_code(code)]) if code is not None else None
        record = replace(
            record,
            Modality="MG",
            PaddleDescription=paddle,
            ViewModifierCodeSequence=seq,
            ViewPosition=view_pos,
        )
        assert record.is_spot_compression == exp

    @pytest.mark.parametrize(
        "code,paddle,exp",
        [
            pytest.param("spot compression", None, False),
            pytest.param("magnified", None, True),
            pytest.param("magnification", None, True),
            pytest.param(None, "7.5CM SPTMAG", True),
        ],
    )
    def test_is_magnified(self, code, paddle, exp, record_factory):
        rec = record_factory(view_modifier_code=code)
        rec = rec.replace(PaddleDescription=paddle)
        assert rec.is_magnified == exp

    @pytest.mark.parametrize(
        "code,exp",
        [
            pytest.param("implant displaced", True),
            pytest.param("magnified", False),
            pytest.param("", False),
        ],
    )
    def test_is_implant_displaced(self, code, exp, record_factory):
        record = record_factory()
        seq = Sequence([make_view_modifier_code(code)]) if code is not None else None
        record = replace(
            record,
            Modality="MG",
            ViewModifierCodeSequence=seq,
        )
        assert record.is_implant_displaced == exp

    @pytest.mark.parametrize(
        "view_pos,exp",
        [
            pytest.param(ViewPosition.MLO, True),
            pytest.param(ViewPosition.CC, True),
            *[pytest.param(x, False) for x in ViewPosition if x not in (ViewPosition.MLO, ViewPosition.CC)],
        ],
    )
    def test_is_standard_mammo_view(self, view_pos, exp, record_factory):
        record = record_factory(view_position=view_pos)
        assert record.is_standard_mammo_view == exp

    @pytest.mark.parametrize(
        "attr_name,val,exp",
        [
            pytest.param(None, None, True),
            pytest.param("is_spot_compression", True, False),
            pytest.param("is_magnified", True, False),
            # We can't set `is_secondary_capture` directly since it's not a cached_property
            pytest.param("SOPClassUID", SecondaryCaptureImageStorage, False),
            pytest.param("is_for_processing", True, False),
            pytest.param("is_cad", True, False),
            pytest.param("is_stereo", True, False),
            pytest.param("is_infra_mammary_fold", True, False),
            pytest.param("is_nipple_in_profile", True, False),
            pytest.param("is_anterior_compression", True, False),
            pytest.param("is_tangential", True, False),
            # Implant displaced views are standard
            pytest.param("is_implant_displaced", True, True),
            pytest.param("is_specimen", True, False),
        ],
    )
    def test_is_standard_mammo_view_modifiers(self, record_factory, attr_name, val, exp):
        record = record_factory(view_position=ViewPosition.MLO)
        if attr_name is not None:
            assert hasattr(record, attr_name)
            object.__setattr__(record, attr_name, val)
            assert getattr(record, attr_name) == val
        assert record.is_standard_mammo_view == exp

    @pytest.mark.parametrize("secondary_capture", [False, True])
    def test_is_complete_mammo_case(self, secondary_capture, record_factory):
        record = record_factory()
        if secondary_capture:
            record = record.replace(SOPClassUID=SecondaryCaptureImageStorage)
        # should be incomplete until after this loop
        records: List[MammogramFileRecord] = []
        for laterality, view_pos in STANDARD_MAMMO_VIEWS:
            assert not MammogramFileRecord.is_complete_mammo_case(records)
            rec = replace(record, laterality=laterality, view_position=view_pos)
            records.append(rec)

        actual = MammogramFileRecord.is_complete_mammo_case(records)
        expected = not secondary_capture
        assert actual == expected

    @pytest.mark.parametrize(
        "mtype,spot,mag,id,laterality,view_pos,uid,secondary,for_proc,stereo,specimen,exp",
        [
            pytest.param(
                MammogramType.FFDM,
                False,
                False,
                False,
                Laterality.LEFT,
                ViewPosition.MLO,
                "1",
                False,
                False,
                False,
                False,
                "ffdm_lmlo_1.dcm",
            ),
            pytest.param(
                MammogramType.FFDM,
                False,
                False,
                False,
                Laterality.RIGHT,
                ViewPosition.CC,
                "2",
                False,
                False,
                False,
                False,
                "ffdm_rcc_2.dcm",
            ),
            pytest.param(
                MammogramType.SYNTH,
                True,
                True,
                True,
                Laterality.RIGHT,
                ViewPosition.XCCL,
                "1",
                False,
                False,
                False,
                False,
                "synth_rxccl_spot_mag_id_1.dcm",
            ),
            pytest.param(
                MammogramType.FFDM,
                False,
                False,
                False,
                Laterality.UNKNOWN,
                ViewPosition.UNKNOWN,
                "2",
                False,
                False,
                False,
                False,
                "ffdm_2.dcm",
            ),
            pytest.param(
                MammogramType.SYNTH,
                True,
                True,
                True,
                Laterality.RIGHT,
                ViewPosition.XCCL,
                "1",
                True,
                True,
                False,
                False,
                "synth_rxccl_secondary_proc_spot_mag_id_1.dcm",
            ),
            pytest.param(
                MammogramType.FFDM,
                False,
                False,
                False,
                Laterality.RIGHT,
                ViewPosition.CC,
                "1",
                False,
                False,
                True,
                False,
                "ffdm_rcc_stereo_1.dcm",
            ),
            pytest.param(
                MammogramType.FFDM,
                False,
                False,
                False,
                Laterality.RIGHT,
                ViewPosition.CC,
                "1",
                False,
                False,
                False,
                True,
                "ffdm_rcc_specimen_1.dcm",
            ),
        ],
    )
    def test_standardized_filename(
        self,
        mtype,
        spot,
        mag,
        id,
        laterality,
        view_pos,
        uid,
        secondary,
        for_proc,
        stereo,
        exp,
        record_factory,
        specimen,
    ):
        seq = []
        if spot:
            seq.append(make_view_modifier_code("spot compression"))
        if mag:
            seq.append(make_view_modifier_code("magnification"))
        if id:
            seq.append(make_view_modifier_code("implant displaced"))
        seq = Sequence(seq)
        record = record_factory(
            mammogram_type=mtype,
            laterality=laterality,
            view_position=view_pos,
        )
        record = record.replace(
            mammogram_type=mtype,
            ViewModifierCodeSequence=seq,
            laterality=laterality,
            view_position=view_pos,
            SOPClassUID=SecondaryCaptureImageStorage if secondary else None,
            PresentationIntentType="FOR PROCESSING" if for_proc else None,
            PerformedProcedureStepDescription="Stereo, CC" if stereo else None,
            StudyDescription="specimen" if specimen else None,
        )
        actual = record.standardized_filename(uid)
        assert isinstance(actual, StandardizedFilename)
        assert actual == Path(exp)

    @pytest.mark.parametrize(
        "body_part,study,series,force,exp",
        [
            pytest.param("BREAST", "", "", False, "MG"),
            pytest.param("", "MAMMO SCREEN", "", False, "MG"),
            pytest.param("", "", "MAMMO SCREEN", False, "MG"),
            pytest.param("", "", "", True, "MG"),
        ],
    )
    def test_modality_helper(self, tmp_path, body_part, series, study, force, exp):
        dcm = DicomFactory()(
            Modality="CT",
            SeriesDescription=series,
            StudyDescription=study,
            BodyPartExamined=body_part,
        )
        path = Path(tmp_path, "test.dcm")
        dcm.save_as(path)
        helpers = [ModalityHelper(force=force)]
        rec = MammogramFileRecord.from_file(path, helpers=helpers)
        assert rec.Modality == exp

    @pytest.mark.parametrize(
        "l1,l2,exp",
        [
            pytest.param(None, None, Laterality.NONE),
            pytest.param(Laterality.LEFT, None, Laterality.LEFT),
            pytest.param(None, Laterality.RIGHT, Laterality.RIGHT),
            pytest.param(Laterality.LEFT, Laterality.RIGHT, Laterality.BILATERAL),
        ],
    )
    def test_collection_laterality(self, l1, l2, exp, record_factory):
        rec1 = record_factory(laterality=l1)
        rec2 = record_factory(laterality=l2)
        assert MammogramFileRecord.collection_laterality([rec1, rec2]) == exp

    @pytest.mark.parametrize(
        "override",
        ["MG", "US", "CR"],
    )
    def test_from_file_modality_override(self, override, record_factory):
        rec = record_factory(Modality=override)
        path = rec.path
        rec_type = type(rec)
        result = rec_type.from_file(path, overrides={"Modality": "MG"})
        assert result.Modality == "MG"

    @pytest.mark.parametrize(
        "modality",
        ["MG", "US", "CR"],
    )
    def test_from_dicom_modality_override(self, modality, record_factory):
        rec = record_factory()
        path = rec.path
        rec_type = type(rec)

        with pydicom.dcmread(path) as dcm:
            dcm.Modality = modality
            result1 = rec_type.from_dicom(path, pydicom.dcmread(path), overrides={"Modality": "MG"})
            result2 = rec_type.from_dicom(path, pydicom.dcmread(path), Modality="MG")
        assert result1.Modality == "MG"
        assert result2.Modality == "MG"

    @pytest.mark.parametrize(
        "tag,val,exp",
        [
            pytest.param(Tag.StudyDescription, "Stereo, LCC", True),
            pytest.param(Tag.SeriesDescription, "L CC Stereo Projection", True),
            pytest.param(Tag.PerformedProcedureStepDescription, "Stereo, LCC", True),
            pytest.param(None, "", False),
            pytest.param(Tag.SeriesDescription, "MAMMOGRAM, Stereo", True),
            pytest.param(Tag.ImageType, ["DERIVED", "PRIMARY", "STEREO", "RIGHT"], True),
            pytest.param(Tag.ImageType, ["DERIVED", "PRIMARY", "RIGHT"], False),
        ],
    )
    def test_is_stereo(self, tag, val, exp, record_factory):
        record = record_factory()
        if tag is not None:
            record = record.replace(**{tag.name: val})
        assert record.is_stereo == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), False),
            (MammogramFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            # Other is DicomFileRecord, fallback to DicomFileRecord comparison
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                False,
            ),
            # Otherwise compare views
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.ML),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.XCCL),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                False,
            ),
            (
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.FFDM),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.SYNTH),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.SYNTH),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.FFDM),
                False,
            ),
        ],
    )
    def test_lt(self, rec1, rec2, exp):
        assert (rec1 < rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), True),
            (MammogramFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            # Other is DicomFileRecord, fallback to DicomFileRecord comparison
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                True,
            ),
            # Otherwise compare views
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.ML),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.XCCL),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                True,
            ),
            (
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.SYNTH),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.FFDM),
                True,
            ),
        ],
    )
    def test_gt(self, rec1, rec2, exp):
        assert (rec1 > rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), True),
            (MammogramFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), False),
            # Other is DicomFileRecord, fallback to DicomFileRecord comparison
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                True,
            ),
            # Otherwise compare views
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.ML),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.XCCL),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                True,
            ),
            (
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.SYNTH),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.FFDM),
                True,
            ),
        ],
    )
    def test_ge(self, rec1, rec2, exp):
        assert (rec1 >= rec2) == exp

    @pytest.mark.parametrize(
        "rec1,rec2,exp",
        [
            # Other is FileRecord, fallback to path comparison
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            (MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None), FileRecord(Path("bar.dcm")), False),
            (MammogramFileRecord(Path("bar.dcm"), SOPInstanceUID=None), FileRecord(Path("foo.dcm")), True),
            # Other is DicomFileRecord, fallback to DicomFileRecord comparison
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=None),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.4")),
                DicomFileRecord(Path("foo.dcm"), SOPInstanceUID=SOPUID("1.2.3")),
                False,
            ),
            # Otherwise compare views
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.MLO),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.ML),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.XCCL),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                False,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, PaddleDescription="SPOT"),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                False,
            ),
            (
                MammogramFileRecord(
                    Path("foo.dcm"),
                    view_position=ViewPosition.CC,
                    BreastImplantPresent="YES",
                    ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                ),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, BreastImplantPresent="YES"),
                True,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.SYNTH),
                MammogramFileRecord(Path("foo.dcm"), view_position=ViewPosition.CC, mammogram_type=MammogramType.FFDM),
                False,
            ),
        ],
    )
    def test_le(self, rec1, rec2, exp):
        assert (rec1 <= rec2) == exp

    @pytest.mark.parametrize(
        "views,to_check,exp",
        [
            (
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m3.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                    MammogramFileRecord(Path("m4.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
            ),
            (
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m3.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                    MammogramFileRecord(Path("m4.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                MammogramView(Laterality.RIGHT, ViewPosition.CC),
                Path("m4.dcm"),
            ),
            (
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m3.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                ],
                MammogramView(Laterality.RIGHT, ViewPosition.CC),
                None,
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        BreastImplantPresent="YES",
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        BreastImplantPresent="YES",
                        ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="prefer-id-over-non-id-1",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        BreastImplantPresent="YES",
                        ViewModifierCodeSequence=cast(Dataset, DicomFactory.code_sequence("implant displaced")),
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        BreastImplantPresent="YES",
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="prefer-id-over-non-id-2",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        mammogram_type=MammogramType.FFDM,
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        mammogram_type=MammogramType.SYNTH,
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="prefer-ffdm-to-synth-1",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        mammogram_type=MammogramType.SYNTH,
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        mammogram_type=MammogramType.FFDM,
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="prefer-ffdm-to-synth-2",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        PaddleDescription="SPOT",
                    ),
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="prefer-standard-to-nonstandard-1",
            ),
            pytest.param(
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        PaddleDescription="SPOT",
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="prefer-standard-to-nonstandard-2",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        Rows=512,
                        Columns=512,
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        Rows=2048,
                        Columns=2048,
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="prefer-high-res-to-low-res-1",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        Rows=2048,
                        Columns=2048,
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        Rows=512,
                        Columns=512,
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="prefer-high-res-to-low-res-2",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        SOPInstanceUID=SOPUID("1.2.3"),
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        SOPInstanceUID=SOPUID("1.2.4"),
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="sopuid-fallback-1",
            ),
            pytest.param(
                [
                    MammogramFileRecord(
                        Path("m1.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        SOPInstanceUID=SOPUID("1.2.4"),
                    ),
                    MammogramFileRecord(
                        Path("m2.dcm"),
                        laterality=Laterality.LEFT,
                        view_position=ViewPosition.MLO,
                        SOPInstanceUID=SOPUID("1.2.3"),
                    ),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="sopuid-fallback-2",
            ),
            pytest.param(
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.ML),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m1.dcm"),
                id="nonstandard-fallback-1",
            ),
            # We should fall back to MLO-like or CC-like views if MLO / CC are unavailable
            pytest.param(
                [
                    MammogramFileRecord(Path("m2.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.ML),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                Path("m2.dcm"),
                id="nonstandard-fallback-2",
            ),
            pytest.param([], MammogramView(Laterality.LEFT, ViewPosition.MLO), None, id="empty-input"),
            pytest.param(
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.AT),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                None,
                id="no-matching-input",
            ),
            pytest.param(
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.UNKNOWN, view_position=ViewPosition.MLO),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                None,
                id="no-matching-laterality",
            ),
            pytest.param(
                [
                    MammogramFileRecord(Path("m1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.UNKNOWN),
                ],
                MammogramView(Laterality.LEFT, ViewPosition.MLO),
                None,
                id="no-matching-view-position",
            ),
        ],
    )
    def test_get_preferred_views(self, views, to_check, exp):
        result = MammogramFileRecord.get_preferred_views(views)
        if exp is not None:
            act = result[to_check]
            assert act is not None
            assert act.path == exp
        else:
            assert result[to_check] is None

    def test_restore_from_view_codes(self, record_factory):
        record = record_factory(view_modifier_code="spot compression")
        rec_dict = record.to_dict()
        restored = FileRecord.from_dict(rec_dict)
        assert isinstance(restored, MammogramFileRecord)
        assert restored.is_spot_compression

    @pytest.mark.parametrize(
        "src,others,exp",
        [
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                    MammogramFileRecord(Path("foo3.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("foo4.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                ],
                MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                    MammogramFileRecord(Path("foo3.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("foo4.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                ],
                MammogramFileRecord(Path("foo4.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.MLO),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                None,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.XCCL),
                ],
                None,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.XCCL),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.UNKNOWN, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.XCCL),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                None,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.BILATERAL, view_position=ViewPosition.CC),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.XCCL),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                None,
            ),
            (
                MammogramFileRecord(Path("foo.dcm"), laterality=Laterality.LEFT, view_position=ViewPosition.UNKNOWN),
                [
                    MammogramFileRecord(Path("foo1.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.XCCL),
                    MammogramFileRecord(Path("foo2.dcm"), laterality=Laterality.RIGHT, view_position=ViewPosition.CC),
                ],
                None,
            ),
        ],
    )
    def test_get_opposing_laterality(self, src, others, exp):
        assert src.get_opposing_laterality(others) == exp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Iterable, Tuple, Union

from pydicom.tag import Tag as PydicomTag

from ._tag_enum import Tag


def create_tag(val: Union[str, int, Tuple[int, int]]) -> Tag:
    r"""Create a tag from a string keyword or int value"""
    if isinstance(val, str):
        try:
            return getattr(Tag, val)
        except AttributeError:
            raise ValueError(f"Invalid tag {val}")
    elif isinstance(val, int):
        return Tag(val)
    elif isinstance(val, tuple):
        return Tag(PydicomTag(*val))
    else:
        raise TypeError(f"Expected int, str or 2-tuple - found {type(val)}")


def get_display_width(tags: Iterable[Tag]) -> int:
    r"""Returns the width of the longest tag string"""
    return max((len(str(tag)) for tag in tags), default=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import sys

import pytest


@pytest.fixture
def dicom_file():
    pydicom = pytest.importorskip("pydicom")
    return pydicom.data.get_testdata_file("CT_small.dcm")


@pytest.mark.parametrize("tag", ["StudyInstanceUID", "PixelData"])
@pytest.mark.parametrize("decompress", [False, True])
def test_hash(dicom_file, capsys, tag, decompress):
    sys.argv = [
        sys.argv[0],
        str(dicom_file),
        "--tag",
        tag,
    ]
    if decompress:
        sys.argv.append("--decompress")
    runpy.run_module("dicom_utils.cli.hash", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert captured.out.startswith("path,hash")
    assert "CT_small.dcm," in captured.out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Set

from ..tags import Tag
from .hash import compute_hash


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path1", help="first path to search")
    parser.add_argument("path2", help="second path to search")
    parser.add_argument("--name", "-n", default="*", help="glob pattern for filename")
    parser.add_argument("--parents", "-p", default=True, action="store_true", help="return unique parent directories")
    parser.add_argument("--jobs", "-j", default=4, help="parallelism")
    parser.add_argument("--tag", "-t", default="PixelData", help="Tag to check overlap on")
    parser.add_argument(
        "--decompress",
        "-d",
        default=False,
        action="store_true",
        help="When tag=PixelData, decompress the data before hashing",
    )
    # TODO add a field to only return DICOMs with readable image data
    return parser


def main(args: argparse.Namespace) -> None:
    path1 = Path(args.path1)
    path2 = Path(args.path2)
    if not path1.is_dir():
        raise NotADirectoryError(path1)
    if not path2.is_dir():
        raise NotADirectoryError(path2)

    path1_seen: Dict[str, Set[Path]] = {}
    path2_seen: Dict[str, Set[Path]] = {}

    def callback(x: Future):
        result = x.result()
        if result is None:
            return
        path, hashval = result

        if path.is_relative_to(args.path1):
            container = path1_seen
        elif path.is_relative_to(args.path2):
            container = path2_seen
        else:
            raise RuntimeError()

        seen = container.get(hashval, set())
        seen.add(path)
        container[hashval] = seen

    futures: List[Future] = []
    tag = getattr(Tag, args.tag)
    with ThreadPoolExecutor(int(args.jobs)) as tp:
        for match in path1.rglob(args.name):
            f = tp.submit(compute_hash, match, tag=tag, decompress=args.decompress)
            f.add_done_callback(callback)
            futures.append(f)
        for match in path2.rglob(args.name):
            f = tp.submit(compute_hash, match, tag=tag, decompress=args.decompress)
            f.add_done_callback(callback)
            futures.append(f)

    for hashval, pathset1 in path1_seen.items():
        pathset2 = path2_seen.get(hashval, {})

        if not pathset2:
            continue

        for p in pathset1:
            print(f"{p},{hashval}")
        for p in pathset2:
            print(f"{p},{hashval}")


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
from argparse import ArgumentParser
from pathlib import Path
from typing import Final

import pydicom
from pydicom import Dataset


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", help="DICOM file with pixel data to remove")
    parser.add_argument("-o", "--output", help="path for stripped file else original path is used", default=None)
    return parser


def strip_pixel_data(ds: Dataset) -> None:
    del ds.PixelData


def to_stripped_dicom(src: Path, dst: Path) -> None:
    ds = pydicom.dcmread(src)
    strip_pixel_data(ds)
    ds.save_as(dst)


def main(args: argparse.Namespace) -> None:
    src_path: Final[Path] = Path(args.path)
    dst_path: Final[Path] = Path(args.output or src_path)
    to_stripped_dicom(src_path, dst_path)


def entrypoint():
    main(get_parser().parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import re
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Final,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from registry import Registry
from tqdm_multiprocessing import ConcurrentMapper

from ..types import MammogramType
from .collection import CollectionHelper, RecordCollection
from .collection import apply_helpers as apply_collection_helpers
from .collection import get_bar_description_suffix
from .input import Input
from .protocols import SupportsGenerated
from .record import HELPER_REGISTRY, FileRecord, MammogramFileRecord, RecordHelper, apply_helpers


YEAR_RE: Final = re.compile(r"\d{4}")

OUTPUT_REGISTRY = Registry("output", bound=Callable[..., "Output"])

R = TypeVar("R", bound=FileRecord)

# For file structure cases/PatientID/Study/file we want
# to select 4 levels from the end to trim leading directories
FILELIST_OFFSET: Final[int] = 4

# For file structure cases/PatientID/Study we want
# to select 3 levels from the end to trim leading directories
CASELIST_OFFSET: Final[int] = 3


@dataclass
class WriteResult:
    key: Tuple[str, ...]
    path: Path
    collection: RecordCollection
    name: str
    metadata: Any = field(repr=False, default=None)

    def __hash__(self) -> int:
        return hash(self.key)


def iterate_symlinks(collection: RecordCollection[R], dest: Path) -> Iterator[Tuple[Path, R]]:
    for pathname, rec in collection.standardized_filenames():
        filepath = Path(dest, pathname)
        yield filepath, rec


class Output(ABC):
    r"""An :class:`Output` implements logic to output a some data product from an :class:`Input`
    or the data products produced by another :class:`Output`.

    Args:
        path:
            Directory the outputs will be written.

        record_filter:
            Filter function to exclude individual :class:`FileRecord`s from the output.

        collection_filter:
            Filter function to exclude entire :class:`RecordCollection`s from the output.

        helpers:
            List of helper names to apply to the output. Can be :class:`RecordHelper` or
            :class:`CollectionHelper` instances.

        use_bar:
            Whether to use a progress bar.

        threads:
            Whether to use threads instead of processes.

        jobs:
            Number of jobs to use. If None, use the number of CPUs.

        chunksize:
            Passed to :class:`ConcurrentMapper`.
    """

    def __init__(
        self,
        root: PathLike,
        output_subdir: PathLike = Path("."),
        record_filter: Optional[Callable[[FileRecord], bool]] = None,
        collection_filter: Optional[Callable[[RecordCollection], bool]] = None,
        helpers: List[str] = [],
        use_bar: bool = True,
        threads: bool = False,
        jobs: Optional[int] = None,
        chunksize: int = 1,
        timeout: Optional[int] = None,
    ):
        self.root = Path(root)
        if not self.root.is_dir():
            raise NotADirectoryError(root)
        self.output_subdir = Path(output_subdir)
        self.path.mkdir(exist_ok=True, parents=True)
        self.record_filter = record_filter
        self.collection_filter = collection_filter
        self.use_bar = use_bar
        self.threads = threads
        self.jobs = jobs
        self.chunksize = chunksize
        self.timeout = timeout
        helper_fns = [HELPER_REGISTRY.get(h).instantiate_with_metadata().fn for h in helpers]
        self.record_helpers: List[RecordHelper] = [h for h in helper_fns if isinstance(h, RecordHelper)]
        self.collection_helpers: List[CollectionHelper] = [h for h in helper_fns if isinstance(h, CollectionHelper)]

    @property
    def path(self) -> Path:
        return self.root / self.output_subdir

    def mapper(self, **kwargs) -> ConcurrentMapper:
        mapper = ConcurrentMapper(self.threads, self.jobs, chunksize=self.chunksize, timeout=self.timeout)
        kwargs.setdefault("disable", not self.use_bar)
        kwargs.setdefault("leave", False)
        mapper.create_bar(**kwargs)
        return mapper

    def __call__(self, inp: Union[Input, Dict[str, Iterable[WriteResult]]]) -> Dict[str, Iterable[WriteResult]]:
        result: Dict[str, Iterable[WriteResult]] = {}
        iterable = inp if isinstance(inp, Input) else inp.items()
        desc = get_bar_description_suffix(self.jobs, self.threads)
        with self.mapper(total=len(iterable), desc=f"Writing {self.__class__.__name__} ({desc})") as mapper:
            it = mapper(
                self._process,
                iterable,
                path=self.path,
                record_filter=self.record_filter,
                collection_filter=self.collection_filter,
            )
            for item in it:
                if item:
                    name = next(x.name for x in item)
                    result[name] = item
        return result

    def _process(
        self,
        kc: Tuple[Sequence[str], RecordCollection],
        path: Path,
        record_filter: Optional[Callable[[FileRecord], bool]] = None,
        collection_filter: Optional[Callable[[RecordCollection], bool]] = None,
    ) -> Optional[Iterable[WriteResult]]:
        key, collection = kc
        name = str(Path(*key))
        # apply record and collection helpers
        if self.record_helpers:
            collection = RecordCollection((apply_helpers(rec.path, rec, self.record_helpers) for rec in collection))
        if self.collection_helpers:
            index = len(key)
            # since we will often use multiple outputs, ensure that the original collection cannot be mutated by
            # passing a copy below
            collection = apply_collection_helpers(copy(collection), self.collection_helpers, index)
            assert isinstance(
                collection, RecordCollection
            ), f"Collection helpers must return a RecordCollection, got {type(collection)}"

        # apply collection and record filters
        if collection_filter is not None and not collection_filter(collection):
            return
        if record_filter is not None:
            collection = collection.filter(record_filter)
            assert isinstance(
                collection, RecordCollection
            ), f"Record filters must return a RecordCollection, got {type(collection)}"

        dest = Path(path, name)
        # pyright fails to recognize this is an abstract classmethod
        written = self.write(key, name, collection, dest)  # type: ignore
        return list(written)

    @abstractmethod
    def write(
        self, key: Tuple[str, ...], name: str, collection: RecordCollection, dest: Path
    ) -> Iterator[WriteResult]: ...


class SymlinkFileOutput(Output):
    def write(self, key: Tuple[str, ...], name: str, collection: RecordCollection, dest: Path) -> Iterator[WriteResult]:
        dest.mkdir(exist_ok=True, parents=True)
        result: RecordCollection[FileRecord] = RecordCollection()
        for filepath, rec in iterate_symlinks(collection, dest):
            assert isinstance(rec, FileRecord)
            generated = isinstance(rec, SupportsGenerated) and rec.generated
            if rec.exists and not generated:
                rec = rec.to_symlink(filepath, overwrite=True)
            result.add(rec)
            yield WriteResult(key, filepath, result, name, None)


class ManifestOutput(Output):
    def __call__(self, inp: Union[Input, Dict[str, Iterable[WriteResult]]]) -> Dict[str, Iterable[WriteResult]]:
        written = super().__call__(inp)
        manifest = self.build_main_manifest(written)
        path = Path(self.root, "manifest.json")
        with open(path, "w") as f:
            json.dump(manifest, f, sort_keys=True, default=str)
        return written

    @classmethod
    def build_main_manifest(cls, values: Dict[str, Iterable[WriteResult]]) -> Dict[str, Any]:
        result = dict()
        for name, write_results in values.items():
            for write_result in write_results:
                container = result
                for k in write_result.key:
                    container = container.setdefault(k, {})
                if write_result.metadata:
                    container.update(write_result.metadata)
        return result

    def write(
        self,
        key: Tuple[str, ...],
        name: str,
        collection: RecordCollection,
        dest: Path,
    ) -> Iterator[WriteResult]:
        # we want the manifest to reflect symlink paths and real paths.
        symlink_collection = RecordCollection()
        for symlink_path, rec in iterate_symlinks(collection, dest):
            symlink_rec = rec.replace(path=Path(symlink_path.absolute()))
            symlink_collection.add(symlink_rec)
        collection = symlink_collection

        metadata = collection.to_dict()
        metadata["name"] = name

        dest.mkdir(exist_ok=True, parents=True)
        path = Path(dest, "manifest.json")
        with open(path, "w") as f:
            json.dump(metadata, f, indent=2, sort_keys=True, default=str)
        yield WriteResult(key, path, collection, name=name, metadata=metadata)


# TODO: consider deprecating this
class FileListOutput(SymlinkFileOutput):
    def __init__(
        self,
        root: PathLike,
        output_subdir: PathLike = Path("."),
        filelist_output_subdir: PathLike = Path("filelist"),
        record_filter: Optional[Callable[[FileRecord], bool]] = None,
        collection_filter: Optional[Callable[[RecordCollection], bool]] = None,
        helpers: List[str] = [],
        use_bar: bool = True,
        threads: bool = False,
        jobs: Optional[int] = None,
        chunksize: int = 8,
        by_case: Optional[bool] = None,
        min_name_len: Optional[int] = 3,
    ):
        super().__init__(
            root, output_subdir, record_filter, collection_filter, helpers, use_bar, threads, jobs, chunksize
        )
        self.filelist_output_subdir = Path(filelist_output_subdir)
        self.by_case = by_case if by_case is not None else (self.record_filter is None)
        self.min_name_len = min_name_len

    @property
    def filelist_path(self) -> Path:
        return self.root / self.filelist_output_subdir

    def write(self, key: Tuple[str, ...], name: str, collection: RecordCollection, dest: Path) -> Iterator[WriteResult]:
        iterator = (
            self._iterate_case_entries(collection, dest)
            if self.by_case
            else self._iterate_file_entries(collection, dest)
        )
        for entry in iterator:
            yield WriteResult(key, dest, collection, name, metadata=entry)

    def _iterate_file_entries(self, collection: RecordCollection, dest: Path) -> Iterator[str]:
        for filepath, rec in iterate_symlinks(collection, dest):
            yield str(self.path_to_filelist_entry(filepath))

    def _iterate_case_entries(self, collection: RecordCollection, dest: Path) -> Iterator[str]:
        seen_entries: Set[Path] = set()
        for file_entry in self._iterate_file_entries(collection, dest):
            entry = Path(file_entry).parent
            if entry in seen_entries:
                continue
            if not self.min_name_len or len(entry.parts) >= self.min_name_len:
                seen_entries.add(entry)
                yield str(entry)

    def __call__(self, inp: Union[Input, Dict[str, Iterable[WriteResult]]]) -> Dict[str, Iterable[WriteResult]]:
        written = super().__call__(inp)
        filelist = self.build_filelist(written)
        dest = Path(self.filelist_path)
        dest.parent.mkdir(exist_ok=True, parents=True)
        with open(dest, "w") as f:
            for name, entries in filelist.items():
                for entry in entries:
                    f.write(f"{entry}\n")
        return written

    def build_filelist(self, values: Dict[str, Iterable[WriteResult]]) -> Dict[str, List[str]]:
        result: Dict[str, List[str]] = {}
        for name, write_results in values.items():
            for write_result in write_results:
                assert isinstance(write_result.metadata, str)
                container = result.setdefault(name, [])
                container.append(write_result.metadata)
        return result

    def path_to_filelist_entry(self, path: Path) -> Path:
        return path.relative_to(self.root)


def is_complete_case(c: RecordCollection) -> bool:
    mammograms = c.filter(lambda rec: isinstance(rec, MammogramFileRecord))
    assert all(isinstance(rec, MammogramFileRecord) for rec in mammograms)
    mammograms = cast(Iterable[MammogramFileRecord], mammograms)
    return MammogramFileRecord.is_complete_mammo_case(mammograms)


def is_mammogram_case(c: RecordCollection) -> bool:
    return any(isinstance(rec, MammogramFileRecord) for rec in c)


def is_mammogram_record(
    rec: FileRecord, mtype: Optional[MammogramType] = None, secondary: bool = False, proc: bool = False
) -> bool:
    return (
        isinstance(rec, MammogramFileRecord)
        and (mtype is None or rec.mammogram_type == mtype)
        and (secondary or not rec.is_secondary_capture)
        and (proc or not rec.is_for_processing)
    )


def is_2d_mammogram(rec: FileRecord) -> bool:
    return isinstance(rec, MammogramFileRecord) and rec.is_2d


def is_spot_mag(rec: FileRecord) -> bool:
    return isinstance(rec, MammogramFileRecord) and (rec.is_spot_compression or rec.is_magnified)


def is_standard_ffdm(
    rec: FileRecord,
    secondary: bool = False,
    proc: bool = False,
) -> bool:
    return (
        isinstance(rec, MammogramFileRecord)
        and (rec.mammogram_type == MammogramType.FFDM)
        and (rec.is_standard_mammo_view)
        and (secondary or not rec.is_secondary_capture)
        and (proc or not rec.is_for_processing)
    )


# register primary output groups
# these should not use a record filter
PRIMARY_OUTPUT_GROUPS = [
    ("cases", None),
    ("mammograms", is_mammogram_case),
    ("complete-mammograms", is_complete_case),
]
for name, collection_filter in PRIMARY_OUTPUT_GROUPS:
    OUTPUT_REGISTRY(
        SymlinkFileOutput,
        name=f"symlink-{name}",
        output_subdir=name,
        collection_filter=collection_filter,
    )
    OUTPUT_REGISTRY(
        FileListOutput,
        name=f"filelist-{name}",
        output_subdir=name,
        filelist_output_subdir=Path(f"file_lists/by_case/{name}.txt"),
        collection_filter=collection_filter,
    )

OUTPUT_REGISTRY(ManifestOutput, name="manifest", output_subdir="cases")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import ClassVar, Type

import pytest

from dicom_utils.container import HELPER_REGISTRY, FileRecord, MammogramFileRecord, RecordHelper
from dicom_utils.container.collection import CollectionHelper
from dicom_utils.container.input import Input
from dicom_utils.container.output import ManifestOutput, Output, SymlinkFileOutput, is_2d_mammogram
from dicom_utils.dicom_factory import CompleteMammographyStudyFactory


@pytest.fixture
def dicom_files(tmp_path):
    paths = []
    for i in range(3):
        fact = CompleteMammographyStudyFactory(seed=i, PatientID=f"Patient{i}", StudyInstanceUID=f"Study{i}")
        for dcm in fact():
            path = (tmp_path / dcm.SOPInstanceUID).with_suffix(".dcm")
            dcm.save_as(path)
            paths.append(path)
    return paths


@HELPER_REGISTRY(name="set-implant")
class SetImplantHelper(RecordHelper):
    def __call__(self, path, record):
        if isinstance(record, MammogramFileRecord):
            record = record.replace(BreastImplantPresent="Y")
        return record


@HELPER_REGISTRY(name="add-new")
class AddNewRecord(CollectionHelper):
    def __call__(self, collection, *args, **kwargs):
        rec = FileRecord(Path("dummy.dcm"))
        collection.add(rec)
        return collection


def is_3d_case(col):
    return any(isinstance(rec, MammogramFileRecord) and rec.is_2d for rec in col)


class BaseOutputTest:
    INPUT_TYPE: ClassVar[Type[Output]]
    DEFAULT_LEN: ClassVar[int] = 3
    DEFAULT_IMAGES: ClassVar[int] = 12

    def test_basic(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest)
        result = output(inp)
        assert len(result) == self.DEFAULT_LEN
        for r in result.values():
            assert len(r) == self.DEFAULT_IMAGES  # type: ignore
            assert all(f.path.is_file() for f in r)

    def test_record_filter(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest, record_filter=is_2d_mammogram)
        result = output(inp)
        assert len(result) == self.DEFAULT_LEN
        for r in result.values():
            for write_result in r:
                assert all(not isinstance(rec, MammogramFileRecord) or rec.is_2d for rec in write_result.collection)

    def test_record_helper(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest, helpers=["set-implant"])
        result = output(inp)
        assert len(result) == self.DEFAULT_LEN
        for r in result.values():
            assert len(r) == self.DEFAULT_IMAGES  # type: ignore
            for write_result in r:
                assert all(not isinstance(rec, MammogramFileRecord) or rec.has_uid for rec in write_result.collection)


class TestSymlinkFileOutput(BaseOutputTest):
    INPUT_TYPE: ClassVar[Type[Output]] = SymlinkFileOutput

    def test_collection_filter(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest, record_filter=is_3d_case)
        result = output(inp)
        assert len(result) == 0

    def test_collection_filter_preserves_collection(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest, record_filter=is_3d_case)
        result = output(inp)
        assert len(result) == 0
        output = self.INPUT_TYPE(dest)
        result = output(inp)
        assert len(result) == self.DEFAULT_LEN

    def test_collection_helper(self, tmp_path, dicom_files):
        source = tmp_path
        dest = Path(tmp_path, "dest")
        dest.mkdir()
        inp = Input(source, use_bar=False)
        output = self.INPUT_TYPE(dest, helpers=["add-new"], threads=True)
        result = output(inp)
        assert len(result) == self.DEFAULT_LEN
        for r in result.values():
            assert len(r) == self.DEFAULT_IMAGES + 1  # type: ignore
            for write_result in r:
                assert any(rec.path.name == "dummy.dcm" for rec in write_result.collection)


class TestManifestOutput(BaseOutputTest):
    INPUT_TYPE: ClassVar[Type[Output]] = ManifestOutput
    DEFAULT_LEN: ClassVar[int] = 3
    DEFAULT_IMAGES: ClassVar[int] = 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from argparse import ArgumentParser, Namespace
from os import PathLike
from pathlib import Path
from typing import Dict, Iterable, Optional, Union

from tqdm import tqdm

from ..container.collection import RecordCollection
from ..container.group import GROUP_REGISTRY
from ..container.input import Input
from ..container.output import OUTPUT_REGISTRY
from ..container.record import HELPER_REGISTRY


def organize(
    sources: Union[PathLike, Iterable[PathLike]],
    dest: PathLike,
    records: Optional[Iterable[str]] = None,
    groups: Iterable[str] = ["patient-id", "study-uid"],
    helpers: Iterable[str] = [],
    namers: Iterable[str] = ["patient-id", "study-uid"],
    outputs: Iterable[str] = ["symlink-cases"],
    use_bar: bool = True,
    jobs: Optional[int] = None,
    threads: bool = False,
    timeout: Optional[int] = None,
    **kwargs,
) -> Dict[str, Dict[str, RecordCollection]]:
    inp = Input(
        sources,
        records,
        groups,
        helpers,
        namers,
        use_bar=use_bar,
        jobs=jobs,
        threads=threads,
        timeout=timeout,
        **kwargs,
    )

    result: Dict[str, Dict[str, RecordCollection]] = {}
    for output_name in tqdm(outputs, desc="Writing outputs", disable=not use_bar):
        output = OUTPUT_REGISTRY.get(output_name).instantiate_with_metadata(
            root=dest,
            use_bar=use_bar,
            jobs=jobs,
            threads=True,
            timeout=timeout,
        )
        result[output_name] = output(inp)

    return result


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("paths", nargs="+", type=Path, help="path to source files")
    parser.add_argument("dest", type=Path, help="path to outputs")
    parser.add_argument(
        "-g",
        "--group",
        default="study_uid",
        choices=GROUP_REGISTRY.available_keys(),
        help="grouping function",
    )

    parser.add_argument(
        "-o",
        "--output",
        nargs="+",
        default=list(OUTPUT_REGISTRY.available_keys()),
        choices=OUTPUT_REGISTRY.available_keys(),
        help="output functions",
    )

    parser.add_argument(
        "--helpers",
        nargs="+",
        default=[],
        choices=HELPER_REGISTRY.available_keys(),
        help="helper functions",
    )

    parser.add_argument("-m", "--modality", default=None, help="modality override")
    parser.add_argument("-j", "--jobs", default=8, type=int, help="number of parallel jobs")
    parser.add_argument(
        "--allow-non-dicom", default=False, action="store_true", help="keep groups that don't include a DICOM file"
    )
    parser.add_argument(
        "-n", "--numbering-start", default=1, type=int, help="start of numbering for output case symlinks"
    )
    parser.add_argument(
        "--is-sfm", default=False, action="store_true", help="target DICOM files are SFM (as opposed to FFDM)"
    )
    return parser


def main(args: Namespace):
    for p in args.paths:
        if not p.is_dir():
            raise NotADirectoryError(p)
    if not args.dest.is_dir():
        raise NotADirectoryError(args.dest)

    organize(
        args.paths,
        args.dest,
        groups=[args.group],
        outputs=args.output,
        helpers=args.helpers,
        jobs=args.jobs,
        modality=args.modality,
        require_dicom=not args.allow_non_dicom,
    )


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



import logging

import pytest
from pydicom.config import logger as pydicom_logger

from dicom_utils.logging import LoggingLevel, logger, set_logging_level


@pytest.mark.parametrize(
    "logging_level, pydicom_logging_level, expected_logging_level, expected_pydicom_logging_level",
    [
        (LoggingLevel.DEBUG, None, LoggingLevel.DEBUG, LoggingLevel.DEBUG),
        (LoggingLevel.INFO, LoggingLevel.CRITICAL, LoggingLevel.INFO, LoggingLevel.CRITICAL),
    ],
)
def test_set_logging_level(
    logging_level, pydicom_logging_level, expected_logging_level, expected_pydicom_logging_level
):
    set_logging_level(logging_level, pydicom_logging_level)
    assert logging.getLevelName(logger.level) == expected_logging_level.value
    assert logging.getLevelName(pydicom_logger.level) == expected_pydicom_logging_level.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import logging
from argparse import ArgumentParser
from hashlib import md5
from pathlib import Path
from typing import Optional, Tuple

import pydicom
from tqdm import tqdm
from tqdm_multiprocessing import ConcurrentMapper

from ..container.collection import iterate_input_path
from ..dicom import decompress as decompress_fn
from ..dicom import has_dicm_prefix
from ..tags import Tag
from ..types import get_value


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.prog = "hash"
    parser.description = "Compute the hash of one or more DICOM files"
    parser.add_argument("target", default="./", help="path to find from")
    parser.add_argument("--tag", "-t", default="PixelData", help="Tag to compute hash on")
    parser.add_argument(
        "--decompress",
        "-d",
        default=False,
        action="store_true",
        help="When tag=PixelData, decompress the data before hashing",
    )
    parser.add_argument("--jobs", "-j", default=4, help="Parallelism")
    return parser


def compute_hash(path: Path, tag: Tag = Tag.PixelData, decompress: bool = False) -> Optional[Tuple[Path, str]]:
    try:
        if not path.is_file() or not has_dicm_prefix(path):
            return
        ds = pydicom.dcmread(path)
        if tag == Tag.PixelData and decompress:
            ds = decompress_fn(ds)
        value = get_value(ds, tag, "")
        return path, md5(value).hexdigest() if isinstance(value, bytes) else md5(value.encode()).hexdigest()
    except Exception as e:
        logging.exception(f"Failed to hash {path}", exc_info=e)


def main(args: argparse.Namespace) -> None:
    src = Path(args.target)
    sources = [p for p in iterate_input_path(src) if p.is_file()]
    tag = getattr(Tag, args.tag)
    with ConcurrentMapper(jobs=int(args.jobs)) as mapper:
        mapper.create_bar("Hashing", total=len(sources))
        tqdm.write("path,hash")
        for result in mapper(compute_hash, sources, tag=tag, decompress=args.decompress):
            if result is not None:
                path, hashval = result
                tqdm.write(f"{path},{hashval}")


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



import copy
from pathlib import Path
from typing import Any, List

import pytest

from dicom_utils.anonymize import *
from dicom_utils.private import MEDCOG_ADDR, MEDCOG_NAME, PRIVATE_ELEMENTS_DESCRIPTION


CRITICAL_PHI_TAGS: Final[List[Tag]] = [
    Tag.InstitutionAddress,
    Tag.OperatorsName,
    Tag.PatientAddress,
    Tag.PatientBirthDate,
    Tag.PatientName,
    Tag.PatientTelephoneNumbers,
    Tag.ReferringPhysicianName,
    Tag.ReferringPhysicianAddress,
    Tag.StudyDate,
]

POSSIBLE_ANON_VALS: Final[List[str]] = ["", "ANONYMIZED"]


@pytest.mark.parametrize(
    "test_data",
    [
        ("1", 1),
        ("078Y", 78),
        ("090Y", 90),
        ("abcdefgh120ijklmnopqrts", 120),
    ],
)
def test_str_to_first_int(test_data) -> None:
    input_string, expected_int = test_data
    assert expected_int == str_to_first_int(input_string)


@pytest.mark.parametrize(
    "test_data",
    [
        ("1", "001Y"),
        ("000078Y", "078Y"),
        ("90Y", "90Y+"),
        ("abcdefgh120ijklmnopqrts", "90Y+"),
    ],
)
def test_anonymize_age(test_data) -> None:
    input_string, expected_output = test_data
    assert expected_output == anonymize_age(input_string)


@pytest.mark.parametrize(
    "actual_age, expected_age",
    [
        ("12", "012Y"),
        ("000052Y", "052Y"),
        ("90Y", "90Y+"),
        ("5642", "90Y+"),
    ],
)
def test_age_anon(test_dicom: pydicom.Dataset, actual_age: str, expected_age: str) -> None:
    ds = copy.deepcopy(test_dicom)
    ds.PatientAge = actual_age

    anonymize(ds)

    assert ds.PatientAge == expected_age


def test_private_tags(test_dicom) -> None:
    medcog_elements = get_medcog_elements(test_dicom)

    ds = copy.deepcopy(test_dicom)
    anonymize(ds)

    block = get_medcog_block(ds)
    assert block[0].value == PRIVATE_ELEMENTS_DESCRIPTION
    for i, element in enumerate(medcog_elements):
        assert block[i + 1].VR == element.VR
        assert block[i + 1].value == element.value


def test_anonymize(test_dicom) -> None:
    # Additional testing is present in the `dicom-anonymizer` repo
    # This test is more of a sanity check
    ds = copy.deepcopy(test_dicom)

    for tag in CRITICAL_PHI_TAGS:
        s = "19000101" if "date" in repr(tag).lower() else "filler string"
        ds.add_new(tag.tag_tuple, "LO", s)

    for tag in CRITICAL_PHI_TAGS:
        assert hasattr(ds, tag.name)
        assert ds[tag] != ""

    anonymize(ds)

    assert ds.PatientID[: len(PATIENT_ID_PREFIX)] == PATIENT_ID_PREFIX

    for tag in CRITICAL_PHI_TAGS:
        assert not hasattr(ds, tag.name) or (ds[tag].value in POSSIBLE_ANON_VALS)


def test_patient_id_anon(test_dicom) -> None:
    datasets = [copy.deepcopy(test_dicom) for _ in range(3)]

    for i in [0, 2]:
        datasets[i].PatientID = "one patient"
    datasets[1].PatientID = "another patient"

    for d in datasets:
        anonymize(d)

    for d in datasets:
        assert d.PatientID[: len(PATIENT_ID_PREFIX)] == PATIENT_ID_PREFIX

    assert datasets[0].PatientID == datasets[2].PatientID
    assert datasets[0].PatientID != datasets[1].PatientID


@pytest.mark.parametrize(
    "actual_descr, expected_descr",
    [
        ("test12345", "test12345"),
    ],
)
def test_deriv_descr_anon(test_dicom: pydicom.Dataset, actual_descr: str, expected_descr: str) -> None:
    ds = copy.deepcopy(test_dicom)
    ds.DerivationDescription = actual_descr

    anonymize(ds)

    assert ds.DerivationDescription == expected_descr, f"{ds.DerivationDescription} != {expected_descr}"


@pytest.mark.parametrize(
    "tag_a, tag_b, filler_str",
    [
        (Tag.PatientName, Tag.InstitutionName, "test123"),
        (Tag.InstitutionName, Tag.PatientName, "test123"),
        (Tag.InstitutionAddress, Tag.PatientName, "1234 Street, City"),
        (Tag.PatientName, Tag.InstitutionAddress, "1234 Street, City"),
        (Tag.PatientAddress, Tag.PatientName, "_"),
        (Tag.PatientName, Tag.PatientAddress, "_"),
    ],
)
def test_override_anon_rules(test_dicom: pydicom.Dataset, tag_a: Tag, tag_b: Tag, filler_str: str) -> None:
    ds = copy.deepcopy(test_dicom)
    for t in [tag_a, tag_b]:
        ds.add_new(t.tag_tuple, "LO", filler_str)
    anonymize(ds, {tag_a: preserve_value})
    assert tag_a in ds and ds[tag_a].value == filler_str
    assert not hasattr(ds, tag_b.name) or (ds[tag_b].value in POSSIBLE_ANON_VALS)


def test_is_anonymized(test_dicom: pydicom.Dataset) -> None:
    not_medcog_name = MEDCOG_NAME + " "
    test_dicom.private_block(MEDCOG_ADDR, not_medcog_name, create=True)
    test_dicom.private_block(MEDCOG_ADDR, not_medcog_name, create=False)  # Check block exists (i.e. no exception)

    # The non-Medcognetics block we just created should not make us think that the case is anonymized
    assert not is_anonymized(test_dicom)
    anonymize(test_dicom)
    assert is_anonymized(test_dicom)

    with pytest.raises(Exception):
        not_medcog_name = MEDCOG_NAME + "  "
        # This should not return the MedCognetics block but should raise an exception that the block doesn't exist
        test_dicom.private_block(MEDCOG_ADDR, not_medcog_name, create=False)


def test_double_anonymization(test_dicom: pydicom.Dataset) -> None:
    anonymize(test_dicom)
    with pytest.raises(AssertionError, match="DICOM file is already anonymized"):
        anonymize(test_dicom)


@pytest.mark.parametrize(
    "num_frames, irrad_event_uid, type_str",
    [
        (None, None, "IS"),
        (None, None, "SQ"),
        (1, 1, "IS"),
    ],
)
def test_fix_bad_fields(
    test_dicom: pydicom.Dataset, num_frames: Any, irrad_event_uid: Any, type_str: str, tmp_path: Path
) -> None:
    ds = copy.deepcopy(test_dicom)

    ds.add_new(Tag.NumberOfFrames.tag_tuple, type_str, num_frames)
    ds.add_new(Tag.IrradiationEventUID.tag_tuple, type_str, irrad_event_uid)

    ds.save_as(tmp_dicom_file := tmp_path / "tmp.dcm")

    ds = pydicom.dcmread(tmp_dicom_file)

    # We are only checking that no exceptions are thrown when these tags are accessed
    ds[Tag.NumberOfFrames]
    ds[Tag.IrradiationEventUID]



import copy
import difflib
import os
from typing import Iterator, List

import pydicom
from colorama import Fore
from pydicom import Dataset
from tqdm import tqdm

from dicom_utils.dicom import has_dicm_prefix

from .anonymize import anonymize


def color_diff(diff):
    for line in diff:
        if line.startswith("+"):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith("-"):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith("?"):
            yield Fore.BLUE + line + Fore.RESET
        else:
            pass


def gen_dicoms(path: str) -> Iterator[str]:
    for root, folders, filenames in os.walk(path):
        for f in filenames:
            filename = os.path.join(root, f)
            if has_dicm_prefix(filename):
                yield filename


def dataset_to_str(ds: Dataset) -> List[str]:
    return str(ds).splitlines(keepends=True)


def find_phi(path: str, overwrite: bool, verbose: bool) -> None:
    filenames = [path] if os.path.isfile(path) else list(gen_dicoms(path))

    for filename in tqdm(filenames):
        ds = pydicom.dcmread(filename)
        str(ds)  # I think this forces evaluation of certain fields. Without this, the anonymizer may throw an error.
        ds_str = dataset_to_str(copy.deepcopy(ds))
        anonymize(ds)

        if verbose:
            print(filename)
            for diff in color_diff(difflib.ndiff(ds_str, dataset_to_str(ds))):
                print(diff)

        if overwrite:
            ds.save_as(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import datetime as dt
import json
import re
from typing import Any, Dict, Final, Optional

from pydicom import DataElement

from .tags import Tag
from .types import Dicom


LINE_PATTERN = re.compile(r"^\((\d{4}), (\d{4})\)\s*(.*\S).*([A-Z]{2}):\s*(.*)$")

GROUPS: Dict[str, str] = {"image": "0028", "media": "0008"}

# fields with values above this limit will be dropped
MAX_FIELD_LENGTH: Final[int] = 100

AGE_TAG = Tag.PatientAge
DOB = Tag.PatientBirthDate
STUDY_DATE = Tag.StudyDate
CONTENT_DATE = Tag.ContentDate
ACQUISITION_DATE = Tag.AcquisitionDate
DENSITY = Tag.Density


def is_inverted(photo_interp: str) -> bool:
    """Checks if pixel value 0 corresponds to white. See DICOM specification for more details."""
    if photo_interp == "MONOCHROME1":
        return True
    elif photo_interp != "MONOCHROME2":
        # I don't think we need to handle any interpretations besides MONOCHROME1
        # and MONOCHROME2 in the case of mammograms.
        raise Exception(f"Unexpected photometric interpretation '{photo_interp}'")
    return False


def drop_empty_tags(dcm: Dicom) -> Dicom:
    r"""Drops empty tags from a DICOM object. This applies only to the in-memory
    DICOM object - no modifications are made to the file on disk.

    Args:
        dcm (:class:`pydicom.FileDataset`):
            The DICOM object to drop empty tags from

    Returns:
        ``dcm`` with all empty tags removed.
    """
    drop_keys = []
    for key in dcm.keys():
        try:
            tag = dcm.get(key)
            if tag.is_empty:
                drop_keys.append(key)
        except AttributeError:
            drop_keys.append(key)
    for k in drop_keys:
        del dcm[k]
    return dcm


def drop_fields_by_length(dcm: Dicom, max_field_length: int = MAX_FIELD_LENGTH, inplace: bool = False) -> Dicom:
    r"""Drops DICOM fields with values with lengths above a threshold, or fields with
    undefined lengths.

    Args:
        dcm (:class:`pydicom.FileDataset`):
            The DICOM object to process

        max_field_length (int):
            Fields above this length will be dropped

        inplace (bool):
            By default, a deep copy of the DICOM object is created. When ``inplace=True``,
            remove fields from the source DICOM object without creating a copy.
    """
    if not inplace:
        dcm = copy.deepcopy(dcm)

    keys = set(dcm.keys())
    for k in keys:
        v = dcm[k]
        if hasattr(v.value, "__len__") and len(v.value) > max_field_length:
            del dcm[k]
        elif v.is_undefined_length:
            del dcm[k]
    return dcm


def add_patient_age(dcm: Dicom, age: Optional[int] = None) -> Optional[int]:
    r"""Attempts to add the specified patient age to a DICOM object.

    An age can be determined using the following order of precedence:
        1. Reading an existing age tag in the DICOM metadata
        2. Computing an age using the date of birth and capture date metadata
        3. Adding the age passed as an ``age`` parameter.

    If ``age`` is given and ``age`` doesn't match the age determined via DICOM
    metadata, the age will not be updated.

    Args:
        dcm (:class:`pydicom.FileDataset`):
            The DICOM object to add an age to

        age (int, optional):
            The age to attempt to add to ``dcm``.

    Returns:
        The determined patient age in years, or ``None`` if no age could be
        determined.
    """
    capture_date = get_date(dcm)

    if AGE_TAG in dcm:
        dcm_age = int(re.sub("[^0-9]", "", dcm[AGE_TAG].value))
    elif DOB in dcm and capture_date is not None:
        dob = dcm[DOB].value
        try:
            dob = dt.datetime.strptime(dob, "%m%d%Y")
            delta = (capture_date - dob).days / 365.25
            dcm_age = round(delta)
        except ValueError:
            dcm_age = age
    else:
        dcm_age = age

    if dcm_age is not None:
        if dcm_age != age and age is not None:
            print(f"Computed age {dcm_age} doesn't match given age {age}. " "Not updating patient age.")
        else:
            dcm[AGE_TAG] = DataElement(AGE_TAG, "AS", f"{dcm_age:03d}Y")
    elif age is not None:
        dcm[AGE_TAG] = DataElement(AGE_TAG, "AS", f"{age:03d}Y")
    return dcm_age


def get_date(dcm: Dicom, fmt: str = "%Y%m%d") -> Optional[dt.datetime]:
    r"""Attempts to parse a collection date from a DICOM object.

    Dates are parsed using the following order of precedence:
        1. CONTENT_DATE tag
        2. ACQUISITION_DATE tag
        3. STUDY_DATE tag

    Returns:
        Parsed date, or ``None`` if no date could be parsed.
    """
    targets = (CONTENT_DATE, ACQUISITION_DATE, STUDY_DATE)
    for target in targets:
        if target in dcm.keys():
            date = dcm[target].value
            date = dt.datetime.strptime(date, fmt)
            return date
    return None


def add_density(dcm: Dicom, density: float) -> str:
    if DENSITY not in dcm.keys():
        dcm[DENSITY] = DataElement(DENSITY, "FL", density)
        return str(density)
    else:
        return str(dcm[DENSITY].value)


def dicom_to_json(dcm: Dicom) -> Dict[str, Any]:
    r"""Converts DICOM metadata into a JSON dictionary."""
    return json.loads(dcm.to_json())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import sys

import pytest


@pytest.fixture
def dicom_file():
    pydicom = pytest.importorskip("pydicom")
    return pydicom.data.get_testdata_file("CT_small.dcm")


def test_cat(dicom_file, capsys):
    sys.argv = [
        sys.argv[0],
        str(dicom_file),
    ]
    runpy.run_module("dicom_utils.cli.cat", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert "CT_small.dcm" in captured.out
    assert "1.3.6.1.4.1.5962.1.2.1.20040119072730.12322" in captured.out


def test_cat_tags(dicom_file, capsys):
    sys.argv = [sys.argv[0], str(dicom_file), "--tags", "StudyInstanceUID"]
    runpy.run_module("dicom_utils.cli.cat", run_name="__main__", alter_sys=True)
    captured = capsys.readouterr()
    assert "1.3.6.1.4.1.5962.1.2.1.20040119072730.12322" in captured.out
    assert "Study Date" not in captured.out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import inspect
from argparse import ArgumentParser, _SubParsersAction
from typing import Callable

from .cli.cat import get_parser as cat_parser
from .cli.cat import main as cat_main
from .cli.decompress import get_parser as decompress_parser
from .cli.decompress import main as decompress_main
from .cli.dicom2img import get_parser as dicom2img_parser
from .cli.dicom2img import main as dicom2img_main
from .cli.dicom_types import get_parser as dicom_types_parser
from .cli.dicom_types import main as dicom_types_main
from .cli.dicomphi import get_parser as dicomphi_parser
from .cli.dicomphi import main as dicomphi_main
from .cli.find import get_parser as find_parser
from .cli.find import main as find_main
from .cli.merge_series import get_parser as merge_series_parser
from .cli.merge_series import main as merge_series_main
from .cli.overlap import get_parser as overlap_parser
from .cli.overlap import main as overlap_main
from .cli.strip import get_parser as strip_parser
from .cli.strip import main as strip_main
from .cli.validate import get_parser as validate_parser
from .cli.validate import main as validate_main
from .logging import LoggingLevel, set_logging_level


Main = Callable[[argparse.Namespace], None]
Modifier = Callable[[ArgumentParser], None]


def assert_name_match(name: str, main: Main, modifier: Modifier) -> None:
    # Enforce a naming convention to reduce chance of bugs
    main_module = inspect.getmodule(main)
    assert main_module
    assert name in (module := main_module.__name__), f"'{name}' not in '{module}'"
    modifier_module = inspect.getmodule(modifier)
    assert modifier_module
    assert name in (module := modifier_module.__name__), f"'{name}' not in '{module}'"


def add_subparser(subparsers: _SubParsersAction, name: str, help: str, main: Main, modifier: Modifier) -> None:
    assert_name_match(name, main, modifier)
    subparser = subparsers.add_parser(name, help=help)
    subparser.set_defaults(main=main)
    modifier(subparser)
    ll = LoggingLevel.list()
    subparser.add_argument("--logging_level", "-ll", help="set logging level", choices=ll, default=LoggingLevel.WARNING)
    subparser.add_argument("--pydicom_logging_level", "-pl", help="set pydicom logging level", choices=ll, default=None)


def main() -> None:
    parser = ArgumentParser(description="DICOM CLI utilities")

    subparsers = parser.add_subparsers(help="Operation modes")
    for name, help, main, modifier in [
        ("cat", "Print DICOM metadata", cat_main, cat_parser),
        ("dicom2img", "Convert DICOM to image file", dicom2img_main, dicom2img_parser),
        ("dicomphi", "Find PHI in DICOMs", dicomphi_main, dicomphi_parser),
        ("find", "Find DICOM files", find_main, find_parser),
        ("strip", "Strip pixel data out of DICOMs", strip_main, strip_parser),
        ("dicom_types", "Summarize image types", dicom_types_main, dicom_types_parser),
        ("overlap", "Check overlap of study UIDs between dirs", overlap_main, overlap_parser),
        ("validate", "Validate metadata of a DICOM file", validate_main, validate_parser),
        ("decompress", "Decompress pixel contents of a DICOM file", decompress_main, decompress_parser),
        ("merge_series", "Merge series DICOMs into a single DICOM", merge_series_main, merge_series_parser),
    ]:
        add_subparser(subparsers, name=name, help=help, main=main, modifier=modifier)

    args = parser.parse_args()

    if hasattr(args, "main"):
        set_logging_level(args.logging_level, args.pydicom_logging_level)
        args.main(args)
    else:
        parser.print_usage()


if __name__ == "__main__":
    main()
